#  OpenKiwi: Open-Source Machine Translation Quality Estimation
#  Copyright (C) 2019 Unbabel <openkiwi@unbabel.com>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

from collections import OrderedDict

import torch
from torch import nn

from kiwi import constants as const
from kiwi.metrics import CorrectMetric, ExpectedErrorMetric, PerplexityMetric
from kiwi.models.model import Model, ModelConfig
from kiwi.models.modules.attention import Attention
from kiwi.models.modules.scorer import MLPScorer
from kiwi.models.utils import apply_packed_sequence, replace_token


class PredictorConfig(ModelConfig):
    def __init__(
        self,
        vocabs,
        hidden_pred=400,
        rnn_layers_pred=3,
        dropout_pred=0.0,
        share_embeddings=False,
        embedding_sizes=0,
        target_embeddings_size=200,
        source_embeddings_size=200,
        out_embeddings_size=200,
        predict_inverse=False,
    ):
        """Predictor Hyperparams.
        """
        super().__init__(vocabs)

        # Vocabulary
        self.target_side = const.TARGET
        self.source_side = const.SOURCE
        self.predict_inverse = predict_inverse
        if self.predict_inverse:
            self.source_side, self.target_side = (
                self.target_side,
                self.source_side,
            )
            self.target_vocab_size, self.source_vocab_size = (
                self.source_vocab_size,
                self.target_vocab_size,
            )

        # Architecture
        self.hidden_pred = hidden_pred
        self.rnn_layers_pred = rnn_layers_pred
        self.dropout_pred = dropout_pred
        self.share_embeddings = share_embeddings
        if embedding_sizes:
            self.target_embeddings_size = embedding_sizes
            self.source_embeddings_size = embedding_sizes
            self.out_embeddings_size = embedding_sizes
        else:
            self.target_embeddings_size = target_embeddings_size
            self.source_embeddings_size = source_embeddings_size
            self.out_embeddings_size = out_embeddings_size


@Model.register_subclass
class Predictor(Model):
    """Bidirectional Conditional Language Model

       Implemented after Kim et al 2017, see:
         http://www.statmt.org/wmt17/pdf/WMT63.pdf
    """

    title = 'PredEst Predictor model (an embedder model)'

    def __init__(self, vocabs, **kwargs):
        """
        Args:
          vocabs: Dictionary Mapping Field Names to Vocabularies.
        kwargs:
          config: A state dict of a PredictorConfig object.
          dropout: LSTM dropout Default 0.0
          hidden_pred: LSTM Hidden Size, default 200
          rnn_layers: Default 3
          embedding_sizes: If set, takes precedence over other embedding params
                           Default 100
          source_embeddings_size: Default 100
          target_embeddings_size: Default 100
          out_embeddings_size: Output softmax embedding. Default 100
          share_embeddings: Tie input and output embeddings for target.
                            Default False
          predict_inverse: Predict from target to source. Default False
        """
        super().__init__(vocabs=vocabs, ConfigCls=PredictorConfig, **kwargs)

        scorer = MLPScorer(
            self.config.hidden_pred * 2, self.config.hidden_pred * 2, layers=2
        )

        self.attention = Attention(scorer)
        self.embedding_source = nn.Embedding(
            self.config.source_vocab_size,
            self.config.source_embeddings_size,
            const.PAD_ID,
        )
        self.embedding_target = nn.Embedding(
            self.config.target_vocab_size,
            self.config.target_embeddings_size,
            const.PAD_ID,
        )
        self.lstm_source = nn.LSTM(
            input_size=self.config.source_embeddings_size,
            hidden_size=self.config.hidden_pred,
            num_layers=self.config.rnn_layers_pred,
            batch_first=True,
            dropout=self.config.dropout_pred,
            bidirectional=True,
        )
        self.forward_target = nn.LSTM(
            input_size=self.config.target_embeddings_size,
            hidden_size=self.config.hidden_pred,
            num_layers=self.config.rnn_layers_pred,
            batch_first=True,
            dropout=self.config.dropout_pred,
            bidirectional=False,
        )
        self.backward_target = nn.LSTM(
            input_size=self.config.target_embeddings_size,
            hidden_size=self.config.hidden_pred,
            num_layers=self.config.rnn_layers_pred,
            batch_first=True,
            dropout=self.config.dropout_pred,
            bidirectional=False,
        )

        self.W1 = self.embedding_target
        if not self.config.share_embeddings:
            self.W1 = nn.Embedding(
                self.config.target_vocab_size,
                self.config.out_embeddings_size,
                const.PAD_ID,
            )
        self.W2 = nn.Parameter(
            torch.zeros(
                self.config.out_embeddings_size, self.config.out_embeddings_size
            )
        )
        self.V = nn.Parameter(
            torch.zeros(
                2 * self.config.target_embeddings_size,
                2 * self.config.out_embeddings_size,
            )
        )
        self.C = nn.Parameter(
            torch.zeros(
                2 * self.config.hidden_pred, 2 * self.config.out_embeddings_size
            )
        )
        self.S = nn.Parameter(
            torch.zeros(
                2 * self.config.hidden_pred, 2 * self.config.out_embeddings_size
            )
        )
        for p in self.parameters():
            if len(p.shape) > 1:
                nn.init.xavier_uniform_(p)

        self._loss = nn.CrossEntropyLoss(
            reduction='sum', ignore_index=const.PAD_ID
        )

    @staticmethod
    def fieldset(*args, **kwargs):
        from kiwi.data.fieldsets.predictor import build_fieldset

        return build_fieldset()

    @staticmethod
    def from_options(vocabs, opts):
        """

        Args:
            vocabs:
            opts:

        Returns:

        """
        model = Predictor(
            vocabs,
            hidden_pred=opts.hidden_pred,
            rnn_layers_pred=opts.rnn_layers_pred,
            dropout_pred=opts.dropout_pred,
            share_embeddings=opts.share_embeddings,
            embedding_sizes=opts.embedding_sizes,
            target_embeddings_size=opts.target_embeddings_size,
            source_embeddings_size=opts.source_embeddings_size,
            out_embeddings_size=opts.out_embeddings_size,
            predict_inverse=opts.predict_inverse,
        )
        return model

    def loss(self, model_out, batch, target_side=None):
        if not target_side:
            target_side = self.config.target_side
        target = getattr(batch, target_side)
        # There are no predictions for first/last element
        target = replace_token(target[:, 1:-1], const.STOP_ID, const.PAD_ID)
        # Predicted Class must be in dim 1 for xentropyloss
        logits = model_out[target_side]
        logits = logits.transpose(1, 2)
        loss = self._loss(logits, target)
        loss_dict = OrderedDict()
        loss_dict[target_side] = loss
        loss_dict[const.LOSS] = loss
        return loss_dict

    def forward(self, batch, source_side=None, target_side=None):
        if not source_side:
            source_side = self.config.source_side
        if not target_side:
            target_side = self.config.target_side

        source = getattr(batch, source_side)
        target = getattr(batch, target_side)
        batch_size, target_len = target.shape[:2]
        # Remove First and Last Element (Start / Stop Tokens)
        source_mask = self.get_mask(batch, source_side)[:, 1:-1]
        source_lengths = source_mask.sum(1)
        target_lengths = self.get_mask(batch, target_side).sum(1)
        source_embeddings = self.embedding_source(source)
        target_embeddings = self.embedding_target(target)
        # Source Encoding
        source_contexts, hidden = apply_packed_sequence(
            self.lstm_source, source_embeddings, source_lengths
        )
        # Target Encoding.
        h_forward, h_backward = self._split_hidden(hidden)
        forward_contexts, _ = self.forward_target(target_embeddings, h_forward)
        target_emb_rev = self._reverse_padded_seq(
            target_lengths, target_embeddings
        )
        backward_contexts, _ = self.backward_target(target_emb_rev, h_backward)
        backward_contexts = self._reverse_padded_seq(
            target_lengths, backward_contexts
        )

        # For each position, concatenate left context i-1 and right context i+1
        target_contexts = torch.cat(
            [forward_contexts[:, :-2], backward_contexts[:, 2:]], dim=-1
        )
        # For each position i, concatenate Emeddings i-1 and i+1
        target_embeddings = torch.cat(
            [target_embeddings[:, :-2], target_embeddings[:, 2:]], dim=-1
        )

        # Get Attention vectors for all positions and stack.
        self.attention.set_mask(source_mask.float())
        attns = [
            self.attention(
                target_contexts[:, i], source_contexts, source_contexts
            )
            for i in range(target_len - 2)
        ]
        attns = torch.stack(attns, dim=1)

        # Combine attention, embeddings and target context vectors
        C = torch.einsum('bsi,il->bsl', [attns, self.C])
        V = torch.einsum('bsj,jl->bsl', [target_embeddings, self.V])
        S = torch.einsum('bsk,kl->bsl', [target_contexts, self.S])
        t_tilde = C + V + S
        # Maxout with pooling size 2
        t, _ = torch.max(
            t_tilde.view(
                t_tilde.shape[0], t_tilde.shape[1], t_tilde.shape[-1] // 2, 2
            ),
            dim=-1,
        )

        f = torch.einsum('oh,bso->bsh', [self.W2, t])
        logits = torch.einsum('vh,bsh->bsv', [self.W1.weight, f])
        PreQEFV = torch.einsum('bsh,bsh->bsh', [self.W1(target[:, 1:-1]), f])
        PostQEFV = torch.cat([forward_contexts, backward_contexts], dim=-1)
        return {
            target_side: logits,
            const.PREQEFV: PreQEFV,
            const.POSTQEFV: PostQEFV,
        }

    @staticmethod
    def _reverse_padded_seq(lengths, sequence):
        """ Reverses a batch of padded sequences of different length.
        """
        batch_size, max_length = sequence.shape[:-1]
        reversed_idx = []
        for i in range(batch_size * max_length):
            batch_id = i // max_length
            sent_id = i % max_length
            if sent_id < lengths[batch_id]:
                sent_id_rev = lengths[batch_id] - sent_id - 1
            else:
                sent_id_rev = sent_id  # Padding symbol, don't change order
            reversed_idx.append(max_length * batch_id + sent_id_rev)
        flat_sequence = sequence.contiguous().view(batch_size * max_length, -1)
        reversed_seq = flat_sequence[reversed_idx, :].view(*sequence.shape)
        return reversed_seq

    @staticmethod
    def _split_hidden(hidden):
        """Split Hidden State into forward/backward parts.
        """
        h, c = hidden
        size = h.shape[0]
        idx_forward = torch.arange(0, size, 2, dtype=torch.long)
        idx_backward = torch.arange(1, size, 2, dtype=torch.long)
        hidden_forward = (h[idx_forward], c[idx_forward])
        hidden_backward = (h[idx_backward], c[idx_backward])
        return hidden_forward, hidden_backward

    def metrics(self):
        metrics = []

        main_metric = PerplexityMetric(
            prefix=self.config.target_side,
            target_name=self.config.target_side,
            PAD=const.PAD_ID,
            STOP=const.STOP_ID,
        )
        metrics.append(main_metric)

        metrics.append(
            CorrectMetric(
                prefix=self.config.target_side,
                target_name=self.config.target_side,
                PAD=const.PAD_ID,
                STOP=const.STOP_ID,
            )
        )
        metrics.append(
            ExpectedErrorMetric(
                prefix=self.config.target_side,
                target_name=self.config.target_side,
                PAD=const.PAD_ID,
                STOP=const.STOP_ID,
            )
        )
        return metrics

    def metrics_ordering(self):
        return min



#  OpenKiwi: Open-Source Machine Translation Quality Estimation
#  Copyright (C) 2019 Unbabel <openkiwi@unbabel.com>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import logging

from kiwi.cli.better_argparse import PipelineParser
from kiwi.cli.models import linear, nuqe, predictor_estimator, quetch
from kiwi.lib import predict

logger = logging.getLogger(__name__)


def predict_opts(parser):
    group = parser.add_argument_group("predicting")
    group.add_argument(
        "--batch-size",
        type=int,
        default=64,
        help="Maximum batch size for predicting.",
    )


def build_parser():
    return PipelineParser(
        name="predict",
        model_parsers=[
            nuqe.parser_for_pipeline("predict"),
            predictor_estimator.parser_for_pipeline("predict"),
            quetch.parser_for_pipeline("predict"),
            linear.parser_for_pipeline("predict"),
        ],
        options_fn=predict_opts,
    )


def main(argv=None):
    parser = build_parser()
    options = parser.parse(args=argv)
    # is this needed?
    if options is None:
        return
    predict.predict_from_options(options)


if __name__ == "__main__":  # pragma: no cover
    main()  # pragma: no cover



#  OpenKiwi: Open-Source Machine Translation Quality Estimation
#  Copyright (C) 2019 Unbabel <openkiwi@unbabel.com>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import logging
from collections import OrderedDict

import torch
from torch import nn
from torch.distributions.normal import Normal

from kiwi import constants as const
from kiwi.metrics import (
    CorrectMetric,
    ExpectedErrorMetric,
    F1Metric,
    LogMetric,
    PearsonMetric,
    PerplexityMetric,
    RMSEMetric,
    SpearmanMetric,
    ThresholdCalibrationMetric,
    TokenMetric,
)
from kiwi.models.model import Model
from kiwi.models.predictor import Predictor, PredictorConfig
from kiwi.models.utils import apply_packed_sequence, make_loss_weights

logger = logging.getLogger(__name__)


class EstimatorConfig(PredictorConfig):
    def __init__(
        self,
        vocabs,
        hidden_est=100,
        rnn_layers_est=1,
        mlp_est=True,
        dropout_est=0.0,
        start_stop=False,
        predict_target=True,
        predict_gaps=False,
        predict_source=False,
        token_level=True,
        sentence_level=True,
        sentence_ll=True,
        binary_level=True,
        target_bad_weight=2.0,
        source_bad_weight=2.0,
        gaps_bad_weight=2.0,
        **kwargs
    ):
        """Predictor Estimator Hyperparams.
        """
        super().__init__(vocabs, **kwargs)
        self.start_stop = start_stop or predict_gaps
        self.hidden_est = hidden_est
        self.rnn_layers_est = rnn_layers_est
        self.mlp_est = mlp_est
        self.dropout_est = dropout_est
        self.predict_target = predict_target
        self.predict_gaps = predict_gaps
        self.predict_source = predict_source
        self.token_level = token_level
        self.sentence_level = sentence_level
        self.sentence_ll = sentence_ll
        self.binary_level = binary_level
        self.target_bad_weight = target_bad_weight
        self.source_bad_weight = source_bad_weight
        self.gaps_bad_weight = gaps_bad_weight


@Model.register_subclass
class Estimator(Model):
    title = 'PredEst (Predictor-Estimator)'

    def __init__(
        self, vocabs, predictor_tgt=None, predictor_src=None, **kwargs
    ):

        super().__init__(vocabs=vocabs, ConfigCls=EstimatorConfig, **kwargs)

        if predictor_src:
            self.config.update(predictor_src.config)
        elif predictor_tgt:
            self.config.update(predictor_tgt.config)

        # Predictor Settings #
        predict_tgt = (
            self.config.predict_target
            or self.config.predict_gaps
            or self.config.sentence_level
        )
        if predict_tgt and not predictor_tgt:
            predictor_tgt = Predictor(
                vocabs=vocabs,
                predict_inverse=False,
                hidden_pred=self.config.hidden_pred,
                rnn_layers_pred=self.config.rnn_layers_pred,
                dropout_pred=self.config.dropout_pred,
                target_embeddings_size=self.config.target_embeddings_size,
                source_embeddings_size=self.config.source_embeddings_size,
                out_embeddings_size=self.config.out_embeddings_size,
            )
        if self.config.predict_source and not predictor_src:
            predictor_src = Predictor(
                vocabs=vocabs,
                predict_inverse=True,
                hidden_pred=self.config.hidden_pred,
                rnn_layers_pred=self.config.rnn_layers_pred,
                dropout_pred=self.config.dropout_pred,
                target_embeddings_size=self.config.target_embeddings_size,
                source_embeddings_size=self.config.source_embeddings_size,
                out_embeddings_size=self.config.out_embeddings_size,
            )

        # Update the predictor vocabs if token level == True
        # Required by `get_mask` call in predictor forward with `pe` side
        # to determine padding IDs.
        if self.config.token_level:
            if predictor_src:
                predictor_src.vocabs = vocabs
            if predictor_tgt:
                predictor_tgt.vocabs = vocabs

        self.predictor_tgt = predictor_tgt
        self.predictor_src = predictor_src

        predictor_hidden = self.config.hidden_pred
        embedding_size = self.config.out_embeddings_size
        input_size = 2 * predictor_hidden + embedding_size

        self.nb_classes = len(const.LABELS)
        self.lstm_input_size = input_size

        self.mlp = None
        self.sentence_pred = None
        self.sentence_sigma = None
        self.binary_pred = None
        self.binary_scale = None

        # Build Model #

        if self.config.start_stop:
            self.start_PreQEFV = nn.Parameter(torch.zeros(1, 1, embedding_size))
            self.end_PreQEFV = nn.Parameter(torch.zeros(1, 1, embedding_size))

        if self.config.mlp_est:
            self.mlp = nn.Sequential(
                nn.Linear(input_size, self.config.hidden_est), nn.Tanh()
            )
            self.lstm_input_size = self.config.hidden_est

        self.lstm = nn.LSTM(
            input_size=self.lstm_input_size,
            hidden_size=self.config.hidden_est,
            num_layers=self.config.rnn_layers_est,
            batch_first=True,
            dropout=self.config.dropout_est,
            bidirectional=True,
        )
        self.embedding_out = nn.Linear(
            2 * self.config.hidden_est, self.nb_classes
        )
        if self.config.predict_gaps:
            self.embedding_out_gaps = nn.Linear(
                4 * self.config.hidden_est, self.nb_classes
            )
        self.dropout = None
        if self.config.dropout_est:
            self.dropout = nn.Dropout(self.config.dropout_est)

        # Multitask Learning Objectives #
        sentence_input_size = (
            2 * self.config.rnn_layers_est * self.config.hidden_est
        )
        if self.config.sentence_level:
            self.sentence_pred = nn.Sequential(
                nn.Linear(sentence_input_size, sentence_input_size // 2),
                nn.Sigmoid(),
                nn.Linear(sentence_input_size // 2, sentence_input_size // 4),
                nn.Sigmoid(),
                nn.Linear(sentence_input_size // 4, 1),
            )
            self.sentence_sigma = None
            if self.config.sentence_ll:
                # Predict truncated Gaussian distribution
                self.sentence_sigma = nn.Sequential(
                    nn.Linear(sentence_input_size, sentence_input_size // 2),
                    nn.Sigmoid(),
                    nn.Linear(
                        sentence_input_size // 2, sentence_input_size // 4
                    ),
                    nn.Sigmoid(),
                    nn.Linear(sentence_input_size // 4, 1),
                    nn.Sigmoid(),
                )
        if self.config.binary_level:
            self.binary_pred = nn.Sequential(
                nn.Linear(sentence_input_size, sentence_input_size // 2),
                nn.Tanh(),
                nn.Linear(sentence_input_size // 2, sentence_input_size // 4),
                nn.Tanh(),
                nn.Linear(sentence_input_size // 4, 2),
            )

        # Build Losses #

        # FIXME: Remove dependency on magic numbers
        self.xents = nn.ModuleDict()
        weight = make_loss_weights(
            self.nb_classes, const.BAD_ID, self.config.target_bad_weight
        )

        self.xents[const.TARGET_TAGS] = nn.CrossEntropyLoss(
            reduction='sum', ignore_index=const.PAD_TAGS_ID, weight=weight
        )
        if self.config.predict_source:
            weight = make_loss_weights(
                self.nb_classes, const.BAD_ID, self.config.source_bad_weight
            )
            self.xents[const.SOURCE_TAGS] = nn.CrossEntropyLoss(
                reduction='sum', ignore_index=const.PAD_TAGS_ID, weight=weight
            )
        if self.config.predict_gaps:
            weight = make_loss_weights(
                self.nb_classes, const.BAD_ID, self.config.gaps_bad_weight
            )
            self.xents[const.GAP_TAGS] = nn.CrossEntropyLoss(
                reduction='sum', ignore_index=const.PAD_TAGS_ID, weight=weight
            )
        if self.config.sentence_level and not self.config.sentence_ll:
            self.mse_loss = nn.MSELoss(reduction='sum')
        if self.config.binary_level:
            self.xent_binary = nn.CrossEntropyLoss(reduction='sum')

    @staticmethod
    def fieldset(*args, **kwargs):
        from kiwi.data.fieldsets.predictor_estimator import build_fieldset

        return build_fieldset(*args, **kwargs)

    @staticmethod
    def from_options(vocabs, opts):
        """

        Args:
            vocabs:
            opts:
                predict_target (bool): Predict target tags
                predict_source (bool): Predict source tags
                predict_gaps (bool): Predict gap tags
                token_level (bool): Train predictor using PE field.
                sentence_level (bool): Predict Sentence Scores
                sentence_ll (bool): Use likelihood loss for sentence scores
                                    (instead of squared error)
                binary_level: Predict binary sentence labels
                target_bad_weight: Weight for target tags bad class. Default 3.0
                source_bad_weight: Weight for source tags bad class. Default 3.0
                gaps_bad_weight: Weight for gap tags bad class. Default 3.0

        Returns:

        """
        predictor_src = predictor_tgt = None
        if opts.load_pred_source:
            predictor_src = Predictor.from_file(opts.load_pred_source)
        if opts.load_pred_target:
            predictor_tgt = Predictor.from_file(opts.load_pred_target)

        model = Estimator(
            vocabs,
            predictor_tgt=predictor_tgt,
            predictor_src=predictor_src,
            hidden_est=opts.hidden_est,
            rnn_layers_est=opts.rnn_layers_est,
            mlp_est=opts.mlp_est,
            dropout_est=opts.dropout_est,
            start_stop=opts.start_stop,
            predict_target=opts.predict_target,
            predict_gaps=opts.predict_gaps,
            predict_source=opts.predict_source,
            token_level=opts.token_level,
            sentence_level=opts.sentence_level,
            sentence_ll=opts.sentence_ll,
            binary_level=opts.binary_level,
            target_bad_weight=opts.target_bad_weight,
            source_bad_weight=opts.source_bad_weight,
            gaps_bad_weight=opts.gaps_bad_weight,
            hidden_pred=opts.hidden_pred,
            rnn_layers_pred=opts.rnn_layers_pred,
            dropout_pred=opts.dropout_pred,
            share_embeddings=opts.dropout_est,
            embedding_sizes=opts.embedding_sizes,
            target_embeddings_size=opts.target_embeddings_size,
            source_embeddings_size=opts.source_embeddings_size,
            out_embeddings_size=opts.out_embeddings_size,
            predict_inverse=opts.predict_inverse,
        )
        return model

    def forward(self, batch):
        outputs = OrderedDict()
        contexts_tgt, h_tgt = None, None
        contexts_src, h_src = None, None
        if (
            self.config.predict_target
            or self.config.predict_gaps
            or self.config.sentence_level
        ):
            model_out_tgt = self.predictor_tgt(batch)
            input_seq, target_lengths = self.make_input(
                model_out_tgt, batch, const.TARGET_TAGS
            )

            contexts_tgt, h_tgt = apply_packed_sequence(
                self.lstm, input_seq, target_lengths
            )
            if self.config.predict_target:
                logits = self.predict_tags(contexts_tgt)
                if self.config.start_stop:
                    logits = logits[:, 1:-1]
                outputs[const.TARGET_TAGS] = logits

            if self.config.predict_gaps:
                contexts_gaps = self.make_contexts_gaps(contexts_tgt)
                logits = self.predict_tags(
                    contexts_gaps, out_embed=self.embedding_out_gaps
                )
                outputs[const.GAP_TAGS] = logits
        if self.config.predict_source:
            model_out_src = self.predictor_src(batch)
            input_seq, target_lengths = self.make_input(
                model_out_src, batch, const.SOURCE_TAGS
            )
            contexts_src, h_src = apply_packed_sequence(
                self.lstm, input_seq, target_lengths
            )

            logits = self.predict_tags(contexts_src)
            outputs[const.SOURCE_TAGS] = logits

        # Sentence/Binary/Token Level prediction
        sentence_input = self.make_sentence_input(h_tgt, h_src)
        if self.config.sentence_level:
            outputs.update(self.predict_sentence(sentence_input))

        if self.config.binary_level:
            bin_logits = self.binary_pred(sentence_input).squeeze()
            outputs[const.BINARY] = bin_logits

        if self.config.token_level and hasattr(batch, const.PE):
            if self.predictor_tgt:
                model_out = self.predictor_tgt(batch, target_side=const.PE)
                logits = model_out[const.PE]
                outputs[const.PE] = logits
            if self.predictor_src:
                model_out = self.predictor_src(batch, source_side=const.PE)
                logits = model_out[const.SOURCE]
                outputs[const.SOURCE] = logits

        # TODO remove?
        # if self.use_probs:
        #     logits -= logits.mean(-1, keepdim=True)
        #     logits_exp = logits.exp()
        #     logprobs = logits - logits_exp.sum(-1, keepdim=True).log()
        #     sentence_scores = ((logprobs.exp() * token_mask).sum(1)
        #                        / target_lengths)
        #     sentence_scores = sentence_scores[..., 1 - self.BAD_ID]
        #     binary_logits = (logprobs * token_mask).sum(1)

        return outputs

    def make_input(self, model_out, batch, tagset):
        """Make Input Sequence from predictor outputs. """
        PreQEFV = model_out[const.PREQEFV]
        PostQEFV = model_out[const.POSTQEFV]
        side = const.TARGET
        if tagset == const.SOURCE_TAGS:
            side = const.SOURCE
        token_mask = self.get_mask(batch, side)
        batch_size = token_mask.shape[0]
        target_lengths = token_mask.sum(1)
        if self.config.start_stop:
            target_lengths += 2
            start = self.start_PreQEFV.expand(
                batch_size, 1, self.config.out_embeddings_size
            )
            end = self.end_PreQEFV.expand(
                batch_size, 1, self.config.out_embeddings_size
            )
            PreQEFV = torch.cat((start, PreQEFV, end), dim=1)
        else:
            PostQEFV = PostQEFV[:, 1:-1]

        input_seq = torch.cat([PreQEFV, PostQEFV], dim=-1)
        length, input_dim = input_seq.shape[1:]
        if self.mlp:
            input_flat = input_seq.view(batch_size * length, input_dim)
            input_flat = self.mlp(input_flat)
            input_seq = input_flat.view(
                batch_size, length, self.lstm_input_size
            )
        return input_seq, target_lengths

    def make_contexts_gaps(self, contexts):
        # Concat Contexts Shifted
        contexts = torch.cat((contexts[:, :-1], contexts[:, 1:]), dim=-1)
        return contexts

    def make_sentence_input(self, h_tgt, h_src):
        """Reshape last hidden state. """
        h = h_tgt[0] if h_tgt else h_src[0]
        h = h.contiguous().transpose(0, 1)
        return h.reshape(h.shape[0], -1)

    def predict_sentence(self, sentence_input):
        """Compute Sentence Score predictions."""
        outputs = OrderedDict()
        sentence_scores = self.sentence_pred(sentence_input).squeeze()
        outputs[const.SENTENCE_SCORES] = sentence_scores
        if self.sentence_sigma:
            # Predict truncated Gaussian on [0,1]
            sigma = self.sentence_sigma(sentence_input).squeeze()
            outputs[const.SENT_SIGMA] = sigma
            outputs['SENT_MU'] = outputs[const.SENTENCE_SCORES]
            mean = outputs['SENT_MU'].clone().detach()
            # Compute log-likelihood of x given mu, sigma
            normal = Normal(mean, sigma)
            # Renormalize on [0,1] for truncated Gaussian
            partition_function = (normal.cdf(1) - normal.cdf(0)).detach()
            outputs[const.SENTENCE_SCORES] = mean + (
                (
                    sigma ** 2
                    * (normal.log_prob(0).exp() - normal.log_prob(1).exp())
                )
                / partition_function
            )

        return outputs

    def predict_tags(self, contexts, out_embed=None):
        """Compute Tag Predictions."""
        if not out_embed:
            out_embed = self.embedding_out
        batch_size, length, hidden = contexts.shape
        if self.dropout:
            contexts = self.dropout(contexts)
        # Fold sequence length in batch dimension
        contexts_flat = contexts.contiguous().view(-1, hidden)
        logits_flat = out_embed(contexts_flat)
        logits = logits_flat.view(batch_size, length, self.nb_classes)
        return logits

    def sentence_loss(self, model_out, batch):
        """Compute Sentence score loss"""
        sentence_pred = model_out[const.SENTENCE_SCORES]
        sentence_scores = batch.sentence_scores
        if not self.sentence_sigma:
            return self.mse_loss(sentence_pred, sentence_scores)
        else:
            sigma = model_out[const.SENT_SIGMA]
            mean = model_out['SENT_MU']
            # Compute log-likelihood of x given mu, sigma
            normal = Normal(mean, sigma)
            # Renormalize on [0,1] for truncated Gaussian
            partition_function = (normal.cdf(1) - normal.cdf(0)).detach()
            nll = partition_function.log() - normal.log_prob(sentence_scores)
            return nll.sum()

    def word_loss(self, model_out, batch):
        """Compute Sequence Tagging Loss"""
        word_loss = OrderedDict()
        for tag in const.TAGS:
            if tag in model_out:
                logits = model_out[tag]
                logits = logits.transpose(1, 2)
                word_loss[tag] = self.xents[tag](logits, getattr(batch, tag))
        return word_loss

    def binary_loss(self, model_out, batch):
        """Compute Sentence Classification Loss"""
        labels = getattr(batch, const.BINARY)
        loss = self.xent_binary(model_out[const.BINARY], labels.long())
        return loss

    def loss(self, model_out, batch):
        """Compute Model Loss"""
        loss_dict = self.word_loss(model_out, batch)
        if self.config.sentence_level:
            loss_sent = self.sentence_loss(model_out, batch)
            loss_dict[const.SENTENCE_SCORES] = loss_sent
        if self.config.binary_level:
            loss_bin = self.binary_loss(model_out, batch)
            loss_dict[const.BINARY] = loss_bin

        if const.PE in model_out:
            loss_token = self.predictor_tgt.loss(
                model_out, batch, target_side=const.PE
            )
            loss_dict[const.PE] = loss_token[const.PE]
        if const.SOURCE in model_out:
            loss_token = self.predictor_src.loss(model_out, batch)
            loss_dict[const.SOURCE] = loss_token[const.SOURCE]

        loss_dict[const.LOSS] = sum(loss.sum() for _, loss in loss_dict.items())
        return loss_dict

    def metrics(self):
        metrics = []

        if self.config.predict_target:
            metrics.append(
                F1Metric(
                    prefix=const.TARGET_TAGS,
                    target_name=const.TARGET_TAGS,
                    PAD=const.PAD_TAGS_ID,
                    labels=const.LABELS,
                )
            )
            metrics.append(
                ThresholdCalibrationMetric(
                    prefix=const.TARGET_TAGS,
                    target_name=const.TARGET_TAGS,
                    PAD=const.PAD_TAGS_ID,
                )
            )
            metrics.append(
                CorrectMetric(
                    prefix=const.TARGET_TAGS,
                    target_name=const.TARGET_TAGS,
                    PAD=const.PAD_TAGS_ID,
                )
            )

        if self.config.predict_source:
            metrics.append(
                F1Metric(
                    prefix=const.SOURCE_TAGS,
                    target_name=const.SOURCE_TAGS,
                    PAD=const.PAD_TAGS_ID,
                    labels=const.LABELS,
                )
            )
            metrics.append(
                CorrectMetric(
                    prefix=const.SOURCE_TAGS,
                    target_name=const.SOURCE_TAGS,
                    PAD=const.PAD_TAGS_ID,
                )
            )
        if self.config.predict_gaps:
            metrics.append(
                F1Metric(
                    prefix=const.GAP_TAGS,
                    target_name=const.GAP_TAGS,
                    PAD=const.PAD_TAGS_ID,
                    labels=const.LABELS,
                )
            )
            metrics.append(
                CorrectMetric(
                    prefix=const.GAP_TAGS,
                    target_name=const.GAP_TAGS,
                    PAD=const.PAD_TAGS_ID,
                )
            )

        if self.config.sentence_level:
            metrics.append(RMSEMetric(target_name=const.SENTENCE_SCORES))
            metrics.append(PearsonMetric(target_name=const.SENTENCE_SCORES))
            metrics.append(SpearmanMetric(target_name=const.SENTENCE_SCORES))
            if self.config.sentence_ll:
                metrics.append(
                    LogMetric(targets=[('model_out', const.SENT_SIGMA)])
                )
        if self.config.binary_level:
            metrics.append(
                CorrectMetric(prefix=const.BINARY, target_name=const.BINARY)
            )
        if self.config.token_level and self.predictor_tgt is not None:
            metrics.append(
                CorrectMetric(
                    prefix=const.PE,
                    target_name=const.PE,
                    PAD=const.PAD_ID,
                    STOP=const.STOP_ID,
                )
            )
            metrics.append(
                ExpectedErrorMetric(
                    prefix=const.PE,
                    target_name=const.PE,
                    PAD=const.PAD_ID,
                    STOP=const.STOP_ID,
                )
            )
            metrics.append(
                PerplexityMetric(
                    prefix=const.PE,
                    target_name=const.PE,
                    PAD=const.PAD_ID,
                    STOP=const.STOP_ID,
                )
            )
        if self.config.token_level and self.predictor_src is not None:
            metrics.append(
                CorrectMetric(
                    prefix=const.SOURCE,
                    target_name=const.SOURCE,
                    PAD=const.PAD_ID,
                    STOP=const.STOP_ID,
                )
            )
            metrics.append(
                ExpectedErrorMetric(
                    prefix=const.SOURCE,
                    target_name=const.SOURCE,
                    PAD=const.PAD_ID,
                    STOP=const.STOP_ID,
                )
            )
            metrics.append(
                PerplexityMetric(
                    prefix=const.SOURCE,
                    target_name=const.SOURCE,
                    PAD=const.PAD_ID,
                    STOP=const.STOP_ID,
                )
            )
        metrics.append(
            TokenMetric(
                target_name=const.TARGET, STOP=const.STOP_ID, PAD=const.PAD_ID
            )
        )
        return metrics

    def metrics_ordering(self):
        return max



#  OpenKiwi: Open-Source Machine Translation Quality Estimation
#  Copyright (C) 2019 Unbabel <openkiwi@unbabel.com>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import logging

from kiwi.cli.better_argparse import PipelineParser
from kiwi.cli.opts import PathType
from kiwi.lib import evaluate

logger = logging.getLogger(__name__)


def evaluate_opts(parser):
    # Evaluation options

    group = parser.add_argument_group("Evaluation of WMT Quality Estimation")

    group.add_argument(
        "--type",
        help="Input type for prediction file",
        choices=["probs", "tags"],
        type=str,
        default="probs",
    )
    group.add_argument(
        "--format",
        help="Input format for gold files",
        choices=["wmt17", "wmt18"],
        type=str,
        default="wmt17",
    )
    group.add_argument(
        "--pred-format",
        help="Input format for predicted files. Defaults to the same as "
        "--format.",
        choices=["wmt17", "wmt18"],
        type=str,
        default="wmt18",
    )
    group.add_argument(
        "--sents-avg",
        help="Obtain scores for sentences by averaging over tags or "
        "probabilities.",
        choices=["probs", "tags"],
        type=str,
        # default=None
    )

    # Gold files.
    group.add_argument(
        "--gold-sents",
        help="Sentences gold standard. ",
        type=PathType(exists=True),
        required=False,
    )
    group.add_argument(
        "--gold-target",
        help="Target tags gold standard, or target and gaps "
        'if format == "wmt18".',
        type=PathType(exists=True),
        required=False,
    )
    group.add_argument(
        "--gold-source",
        help="Source tags gold standard.",
        type=PathType(exists=True),
        required=False,
    )
    group.add_argument(
        "--gold-cal",
        help="Target Tags to calibrate.",
        type=PathType(exists=True),
        required=False,
    )

    # Prediction Files
    group.add_argument(
        "--input-dir",
        help="Directory with prediction files generated by predict pipeline. "
        "Setting this argument will evaluate all predictions for "
        "which a gold file is set.",
        nargs="+",
        type=PathType(exists=True),
        # required=True
    )
    group.add_argument(
        "--pred-sents",
        help="Sentences HTER predictions.",
        type=PathType(exists=True),
        nargs="+",
        required=False,
    )
    group.add_argument(
        "--pred-target",
        help="Target predictions; can be tags or probabilities (of BAD). "
        "See --type.",
        type=PathType(exists=True),
        nargs="+",
        required=False,
    )
    group.add_argument(
        "--pred-gaps",
        help="Gap predictions; can be tags or probabilities (of BAD). "
        "(see --type). Use this option for files that only contain gap "
        "tags.",
        type=PathType(exists=True),
        nargs="+",
        required=False,
    )
    group.add_argument(
        "--pred-source",
        help="Source predictions. can be tags or probabilities (of BAD). "
        " See --type.",
        type=PathType(exists=True),
        nargs="+",
        required=False,
    )
    group.add_argument(
        "--pred-cal",
        help="Target Predictions to calibrate.",
        type=PathType(exists=True),
        required=False,
    )


def build_parser():
    return PipelineParser(
        name="evaluate",
        model_parsers=None,
        options_fn=evaluate_opts,
        add_io_options=True,
        add_general_options=False,
        add_logging_options=False,
        add_save_load_options=False,
    )


def main(argv=None):
    parser = build_parser()
    options = parser.parse(args=argv)
    evaluate.evaluate_from_options(options)


if __name__ == "__main__":
    main()



#  OpenKiwi: Open-Source Machine Translation Quality Estimation
#  Copyright (C) 2019 Unbabel <openkiwi@unbabel.com>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import configargparse

from kiwi import __copyright__, __version__
from kiwi.cli.pipelines import evaluate, jackknife, predict, train


def build_parser():
    global parser
    parser = configargparse.get_argument_parser(
        name='main',
        prog='kiwi',
        description='Quality Estimation toolkit',
        add_help=True,
        epilog='Copyright {}'.format(__copyright__),
    )
    parser.add_argument('--version', action='version', version=__version__)
    subparsers = parser.add_subparsers(
        title='Pipelines',
        description="Use 'kiwi <pipeline> (-h | --help)' to check it out.",
        help='Available pipelines:',
        dest='pipeline',
    )
    subparsers.required = True
    subparsers.add_parser(
        'train',
        # parents=[train.parser],
        add_help=False,
        help='Train a QE model',
    )
    subparsers.add_parser(
        'predict',
        # parents=[predict.parser],
        add_help=False,
        help='Use a pre-trained model for prediction',
    )
    subparsers.add_parser(
        'jackknife',
        # parents=[jackknife.parser],
        add_help=False,
        help='Jackknife training data with model',
    )
    subparsers.add_parser(
        'evaluate',
        add_help=False,
        help='Evaluate a model\'s predictions using popular metrics',
    )
    return parser


def cli():
    options, extra_args = build_parser().parse_known_args()

    if options.pipeline == 'train':
        train.main(extra_args)
    if options.pipeline == 'predict':
        predict.main(extra_args)
    # Meta pipelines
    # if options.pipeline == 'search':
    #     search.main(extra_args)
    if options.pipeline == 'jackknife':
        jackknife.main(extra_args)
    if options.pipeline == 'evaluate':
        evaluate.main(extra_args)


if __name__ == '__main__':  # pragma: no cover
    cli()  # pragma: no cover



#  OpenKiwi: Open-Source Machine Translation Quality Estimation
#  Copyright (C) 2019 Unbabel <openkiwi@unbabel.com>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import torch
from torch import nn


class Attention(nn.Module):
    """Generic Attention Implementation.
       Module computes a convex combination of a set of values based on the fit
       of their keys with a query.
    """

    def __init__(self, scorer):
        super().__init__()
        self.scorer = scorer
        self.mask = None

    def forward(self, query, keys, values=None):
        if values is None:
            values = keys
        scores = self.scorer(query, keys)
        # Masked Softmax
        scores = scores - scores.mean(1, keepdim=True)  # numerical stability
        scores = torch.exp(scores)
        if self.mask is not None:
            scores = self.mask * scores
        convex = scores / scores.sum(1, keepdim=True)
        return torch.einsum('bs,bsi->bi', [convex, values])

    def set_mask(self, mask):
        self.mask = mask



"""A generic implementation of a basic trainer."""

#  OpenKiwi: Open-Source Machine Translation Quality Estimation
#  Copyright (C) 2019 Unbabel <openkiwi@unbabel.com>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import logging
from pathlib import Path

import numpy as np

from kiwi import constants as const
from kiwi.models.linear.sparse_vector import SparseVector

from .utils import nearly_eq_tol

logger = logging.getLogger(__name__)


class LinearTrainer(object):
    def __init__(
        self,
        classifier,
        checkpointer,
        algorithm='svm_mira',
        regularization_constant=1e12,
    ):
        self.classifier = classifier
        self.algorithm = algorithm
        self.regularization_constant = regularization_constant
        self.checkpointer = checkpointer
        # Only for training with SGD.
        self.initial_learning_rate = 0.001
        # Only for training with SGD. Change to 'inv' for Pegasos-style
        # updating.
        self.learning_rate_schedule = 'invsqrt'
        # Best metric value (to pick the best iteration).
        self.best_metric_value = -np.inf

    def _make_gradient_step(
        self, parts, features, eta, t, gold_output, predicted_output
    ):
        """Perform a gradient step updating the current model."""
        for r in range(len(parts)):
            if predicted_output[r] == gold_output[r]:
                continue
            if self.classifier.use_binary_features:
                part_features = features[r].to_sparse_vector()
            else:
                part_features = features[r]
            self.classifier.model.make_gradient_step(
                part_features, eta, t, predicted_output[r] - gold_output[r]
            )

    def _make_feature_difference(
        self, parts, features, gold_output, predicted_output
    ):
        """Compute the difference between predicted and gold feature vector."""
        difference = SparseVector()
        for r in range(len(parts)):
            if predicted_output[r] == gold_output[r]:
                continue
            if self.classifier.use_binary_features:
                part_features = features[r].to_sparse_vector()
            else:
                part_features = features[r]
                # FIXME: shouldn't the next line be outside the else?
                difference.add(
                    part_features, predicted_output[r] - gold_output[r]
                )
        return difference

    def run(self, train_iterator, valid_iterator, epochs=50):
        """Train with a general online algorithm."""
        import time

        dataset = self.classifier.create_instances(train_iterator.dataset)
        if not isinstance(valid_iterator, list):
            valid_iterator = [valid_iterator]
        dev_datasets = [
            self.classifier.create_instances(iterator.dataset)
            for iterator in valid_iterator
        ]

        self.classifier.model.clear()
        for epoch in range(epochs):
            tic = time.time()
            logger.info('Epoch %d' % (epoch + 1))
            self._train_epoch(epoch, dataset, dev_datasets)
            toc = time.time()
            logger.info('Elapsed time (epoch): %d' % (toc - tic))
        if self.algorithm != 'svm_sgd':
            self.classifier.model.finalize(len(train_iterator.dataset) * epochs)

        self.checkpointer.check_out()

    def _train_epoch(self, epoch, dataset, dev_datasets):
        """Run one epoch of an online algorithm."""
        algorithm = self.algorithm
        total_loss = 0.0
        total_cost = 0.0
        if algorithm in ['perceptron']:
            num_mistakes = 0
            num_total = 0
        elif algorithm in ['mira', 'svm_mira']:
            truncated = 0
        lambda_coefficient = 1.0 / (
            self.regularization_constant * float(len(dataset))
        )
        t = len(dataset) * epoch

        for instance in dataset:
            # Compute parts, features, and scores.
            parts, gold_output = self.classifier.make_parts(instance)
            features = self.classifier.make_features(instance, parts)
            scores = self.classifier.compute_scores(instance, parts, features)

            # Do the decoding.
            if algorithm in ['perceptron']:
                predicted_output = self.classifier.decoder.decode(
                    instance, parts, scores
                )
                for r in range(len(parts)):
                    num_total += 1
                    if not nearly_eq_tol(
                        gold_output[r], predicted_output[r], 1e-6
                    ):
                        num_mistakes += 1

            elif algorithm in ['mira']:
                predicted_output, cost, loss = self.classifier.decoder.decode_mira(  # NOQA
                    instance, parts, scores, gold_output, True
                )

            elif algorithm in ['svm_mira', 'svm_sgd']:
                predicted_output, cost, loss = self.classifier.decoder.decode_cost_augmented(  # NOQA
                    instance, parts, scores, gold_output
                )
            else:
                raise NotImplementedError

            # Update the total loss and cost.
            if algorithm in ['mira', 'svm_mira', 'svm_sgd']:
                if loss < 0.0:
                    if loss < -1e-12:
                        logger.warning('Negative loss: ' + str(loss))
                    loss = 0.0
                if cost < 0.0:
                    if cost < -1e-12:
                        logger.warning('Negative cost:' + str(cost))
                    cost = 0.0

                total_loss += loss
                total_cost += cost

            num_parts = len(parts)
            assert len(gold_output) == num_parts
            assert len(predicted_output) == num_parts

            # Compute the stepsize.
            if algorithm in ['perceptron']:
                eta = 1.0
            elif algorithm in ['mira', 'svm_mira']:
                difference = self._make_feature_difference(
                    parts, features, gold_output, predicted_output
                )
                squared_norm = difference.squared_norm()
                threshold = 1e-9
                if loss < threshold or squared_norm < threshold:
                    eta = 0.0
                else:
                    eta = loss / squared_norm
                if eta > self.regularization_constant:
                    eta = self.regularization_constant
                    truncated += 1
            elif algorithm in ['svm_sgd']:
                if self.learning_rate_schedule == 'invsqrt':
                    eta = self.initial_learning_rate / np.sqrt(float(t + 1))
                elif self.learning_rate_schedule == 'inv':
                    eta = self.initial_learning_rate / (float(t + 1))
                else:
                    raise NotImplementedError

                # Scale the weight vector.
                decay = 1.0 - eta * lambda_coefficient
                assert decay >= -1e-12
                self.classifier.model.weights.scale(decay)

            # Make gradient step.
            self._make_gradient_step(
                parts, features, eta, t, gold_output, predicted_output
            )

            # Increment the round.
            t += 1

        # Evaluate on development data.
        weights = self.classifier.model.weights.copy()
        averaged_weights = self.classifier.model.averaged_weights.copy()
        if algorithm != 'svm_sgd':
            self.classifier.model.finalize(len(dataset) * (1 + epoch))

        dev_scores = []
        for dev_dataset in dev_datasets:
            predictions = self.classifier.test(dev_dataset)
            dev_score = self.classifier.evaluate(
                dev_dataset, predictions, print_scores=True
            )
            dev_scores.append(dev_score)

        if algorithm in ['perceptron']:
            logger.info(
                '\t'.join(
                    [
                        'Epoch: %d' % (epoch + 1),
                        'Mistakes: %d/%d (%f)'
                        % (
                            num_mistakes,
                            num_total,
                            float(num_mistakes) / float(num_total),
                        ),
                        'Dev scores: %s'
                        % ' '.join(
                            ["%.5g" % (100 * score) for score in dev_scores]
                        ),
                    ]
                )
            )
        else:
            sq_norm = self.classifier.model.weights.squared_norm()
            regularization_value = (
                0.5
                * lambda_coefficient
                * float(len(dataset))
                * weights.squared_norm()
            )
            logger.info(
                '\t'.join(
                    [
                        'Epoch: %d' % (epoch + 1),
                        'Cost: %f' % total_cost,
                        'Loss: %f' % total_loss,
                        'Reg: %f' % regularization_value,
                        'Loss+Reg: %f' % (total_loss + regularization_value),
                        'Norm: %f' % sq_norm,
                        'Dev scores: %s'
                        % ' '.join(
                            ["%.5g" % (100 * score) for score in dev_scores]
                        ),
                    ]
                )
            )

        # If this is the best model so far, save it as the default model.
        # Assume the metric to optimize is on the first dev set, the highest
        # the best.
        # TODO: replace by checkpointer functionality
        metric_value = dev_scores[0]
        if metric_value > self.best_metric_value:
            self.best_metric_value = metric_value
            self.checkpointer.check_in(
                self, self.best_metric_value, epoch=epoch
            )

        self.classifier.model.weights = weights
        self.classifier.model.averaged_weights = averaged_weights

    def save(self, output_directory):
        output_directory = Path(output_directory)
        output_directory.mkdir(exist_ok=True)
        logging.info('Saving training state to {}'.format(output_directory))

        model_path = output_directory / const.MODEL_FILE

        self.classifier.model.save(
            str(model_path), feature_indices=self.classifier.feature_indices
        )
        self.classifier.save(str(model_path))

        return None



#  OpenKiwi: Open-Source Machine Translation Quality Estimation
#  Copyright (C) 2019 Unbabel <openkiwi@unbabel.com>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import numpy as np


class StructuredDecoder(object):
    """An abstract decoder for structured prediction."""

    def __init__(self):
        pass

    def decode(self, instance, parts, scores):
        """Decode, computing the highest-scores output.
        Must return a vector of 0/1 predicted_outputs of the same size
        as parts."""
        raise NotImplementedError

    def decode_mira(
        self, instance, parts, scores, gold_outputs, old_mira=False
    ):
        """Perform cost-augmented decoding or classical MIRA."""
        p = 0.5 - gold_outputs
        q = 0.5 * np.ones(len(gold_outputs)).dot(gold_outputs)
        if old_mira:
            predicted_outputs = self.decode(instance, parts, scores)
        else:
            scores_cost = scores + p
            predicted_outputs = self.decode(instance, parts, scores_cost)
        cost = p.dot(predicted_outputs) + q
        loss = cost + scores.dot(predicted_outputs - gold_outputs)

        return predicted_outputs, cost, loss

    def decode_cost_augmented(self, instance, parts, scores, gold_outputs):
        """Perform cost-augmented decoding."""
        return self.decode_mira(
            instance, parts, scores, gold_outputs, old_mira=False
        )



# -*- coding: utf-8 -*-
"""Several utility functions."""


#  OpenKiwi: Open-Source Machine Translation Quality Estimation
#  Copyright (C) 2019 Unbabel <openkiwi@unbabel.com>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#


def nearly_eq_tol(a, b, tol):
    """Checks if two numbers are equal up to a tolerance."""
    return (a - b) * (a - b) <= tol


def nearly_binary_tol(a, tol):
    """Checks if a number is binary up to a tolerance."""
    return nearly_eq_tol(a, 0.0, tol) or nearly_eq_tol(a, 1.0, tol)


def nearly_zero_tol(a, tol):
    """Checks if a number is zero up to a tolerance."""
    return (a <= tol) and (a >= -tol)



#  OpenKiwi: Open-Source Machine Translation Quality Estimation
#  Copyright (C) 2019 Unbabel <openkiwi@unbabel.com>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import logging

import configargparse
from configargparse import Namespace

from kiwi.cli import opts
from kiwi.cli.opts import PathType
from kiwi.lib.utils import merge_namespaces

logger = logging.getLogger(__name__)


class HyperPipelineParser:
    def __init__(
        self, name, pipeline_parser, pipeline_config_key, options_fn=None
    ):
        self.name = name
        self._pipeline_parser = pipeline_parser
        self._pipeline_config_key = pipeline_config_key.replace('-', '_')

        self._parser = configargparse.get_argument_parser(
            self.name,
            prog='kiwi {}'.format(self.name),
            add_help=False,
            config_file_parser_class=configargparse.YAMLConfigFileParser,
            ignore_unknown_config_file_keys=False,
        )

        self._parser.add(
            '--config',
            required=False,
            is_config_file=True,
            type=PathType(exists=True),
            help='Load config file from path',
        )

        if options_fn is not None:
            options_fn(self._parser)

    def parse(self, args):
        if len(args) == 1 and args[0] in ['-h', '--help']:
            self._parser.print_help()
            return None

        # Parse train pipeline options
        meta_options, extra_args = self._parser.parse_known_args(args)
        print(meta_options)

        if hasattr(meta_options, self._pipeline_config_key):
            extra_args = [
                '--config',
                getattr(meta_options, self._pipeline_config_key),
            ] + extra_args

        pipeline_options = self._pipeline_parser.parse(extra_args)

        options = Namespace()
        options.meta = meta_options
        options.pipeline = pipeline_options

        return options


class PipelineParser:
    _parsers = {}

    def __init__(
        self,
        name,
        model_parsers,
        options_fn=None,
        add_io_options=True,
        add_general_options=True,
        add_logging_options=True,
        add_save_load_options=True,
    ):
        self.name = name

        # Give the option to create pipelines with no models
        if model_parsers is not None:
            self._models = {model.name: model for model in model_parsers}
        else:
            self._models = None

        if name in self._parsers:
            self._parser = self._parsers[name]
        else:
            self._parser = configargparse.get_argument_parser(
                self.name,
                add_help=False,
                prog='kiwi {}'.format(self.name),
                config_file_parser_class=configargparse.YAMLConfigFileParser,
                ignore_unknown_config_file_keys=True,
            )
            self._parsers[name] = self._parser
            self.add_config_option(self._parser)

            if add_io_options:
                opts.io_opts(self._parser)
            if add_general_options:
                opts.general_opts(self._parser)
            if add_logging_options:
                opts.logging_opts(self._parser)
            if add_save_load_options:
                opts.save_load_opts(self._parser)

            if options_fn is not None:
                options_fn(self._parser)

            if model_parsers is not None:
                group = self._parser.add_argument_group('models')
                group.add_argument(
                    '--model',
                    required=True,
                    choices=self._models.keys(),
                    help="Use 'kiwi {} --model <model> --help' for specific "
                    "model options.".format(self.name),
                )

        if 'config' in self._parsers:
            self._config_option_parser = self._parsers['config']
        else:
            self._config_option_parser = configargparse.get_argument_parser(
                'config', add_help=False
            )
            self._parsers['config'] = self._config_option_parser
            self.add_config_option(self._config_option_parser, read_file=False)

    @staticmethod
    def add_config_option(parser, read_file=True):
        parser.add(
            '--config',
            required=False,
            is_config_file=read_file,
            help='Load config file from path',
        )

    def parse_config_file(self, file_name):
        return self.parse(['--config', str(file_name)])

    def parse(self, args):
        if len(args) == 1 and args[0] in ['-h', '--help']:
            self._parser.print_help()
            return None

        # Parse train pipeline options
        pipeline_options, extra_args = self._parser.parse_known_args(args)
        config_option, _ = self._config_option_parser.parse_known_args(args)

        options = Namespace()
        options.pipeline = pipeline_options
        options.model = None
        options.model_api = None

        # Parse specific model options if there are model parsers
        if self._models is not None:
            if pipeline_options.model not in self._models:
                raise KeyError(
                    'Invalid model: {}'.format(pipeline_options.model)
                )

            if config_option:
                extra_args = ['--config', config_option.config] + extra_args

            # Check if there are model parsers
            model_parser = self._models[pipeline_options.model]
            model_options, remaining_args = model_parser.parse_known_args(
                extra_args
            )

            options.model = model_options
            # Retrieve the respective API for the selected model
            options.model_api = model_parser.api_module
        else:
            remaining_args = extra_args

        options.all_options = merge_namespaces(options.pipeline, options.model)

        if remaining_args:
            raise KeyError('Unrecognized options: {}'.format(remaining_args))

        return options


class ModelParser:
    _parsers = {}

    def __init__(self, name, pipeline, options_fn, api_module, title=None):
        self.name = name
        self._title = title
        self._pipeline = pipeline
        self.api_module = api_module

        self._parser = self.get_parser(
            '{}-{}'.format(name, pipeline), description=self._title
        )
        PipelineParser.add_config_option(self._parser)
        options_fn(self._parser)

    @classmethod
    def get_parser(cls, name, **kwargs):
        if name in cls._parsers:
            return cls._parsers[name]
        parser = configargparse.get_argument_parser(
            name,
            prog='... {}'.format(name),
            config_file_parser_class=configargparse.YAMLConfigFileParser,
            ignore_unknown_config_file_keys=True,
            **kwargs,
        )
        cls._parsers[name] = parser
        return parser

    def parse_known_args(self, args):
        return self._parser.parse_known_args(args)



#  OpenKiwi: Open-Source Machine Translation Quality Estimation
#  Copyright (C) 2019 Unbabel <openkiwi@unbabel.com>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import os.path
import warnings
from pathlib import Path

import numpy as np
from more_itertools import flatten
from scipy.stats.stats import pearsonr, rankdata, spearmanr

from kiwi import constants as const
from kiwi.data.utils import read_file
from kiwi.metrics.functions import (
    delta_average,
    f1_scores,
    mean_absolute_error,
    mean_squared_error,
)


def evaluate_from_options(options):
    """
    Evaluates a model's predictions based on the flags received from
     the configuration files. Refer to configuration for a list of
    available configuration flags for the evaluate pipeline.

    Args:
        options (Namespace): Namespace containing all pipeline options
    """

    if options is None:
        return

    setup()
    pipeline_options = options.pipeline

    # flag denoting format so there's no need to always check
    is_wmt18_format = pipeline_options.format.lower() == "wmt18"
    is_wmt18_pred_format = pipeline_options.pred_format.lower() == "wmt18"

    # handling of gold target
    golds = retrieve_gold_standard(pipeline_options, is_wmt18_format)

    # handling of prediction files
    pred_files = retrieve_predictions(pipeline_options, is_wmt18_pred_format)

    if not any(pred_files.values()):
        print(
            "Please specify at least one of these options: "
            "--input-dir, --pred-target, --pred-source, --pred-sents"
        )
        return

    # evaluate word level
    for tag in const.TAGS:
        if tag in golds and pred_files[tag]:
            scores = eval_word_level(golds, pred_files, tag)
            print_scores_table(scores, tag)

    # evaluate sentence level
    if const.SENTENCE_SCORES in golds:
        sent_golds = golds[const.SENTENCE_SCORES]
        sent_preds = retrieve_sentence_predictions(pipeline_options, pred_files)

        if sent_preds:
            sentence_scores, sentence_ranking = eval_sentence_level(
                sent_golds, sent_preds
            )
            print_sentences_scoring_table(sentence_scores)
            print_sentences_ranking_table(sentence_ranking)

        for pred_file in pred_files[const.BINARY]:
            sent_preds.append(pred_file)

    teardown()

    # TODO return some evaluation info besides just printing the graph


def retrieve_gold_standard(pipeline_options, is_wmt18_format):
    golds = {}
    if pipeline_options.gold_target:
        gold_target = _wmt_to_labels(read_file(pipeline_options.gold_target))
        if is_wmt18_format:
            gold_target, gold_gaps = _split_wmt18(gold_target)
            golds[const.GAP_TAGS] = gold_gaps
        golds[const.TARGET_TAGS] = gold_target
    # handling of gold source
    if pipeline_options.gold_source:
        gold_source = _wmt_to_labels(read_file(pipeline_options.gold_source))
        golds[const.SOURCE_TAGS] = gold_source
    # handling of gold sentences
    if pipeline_options.gold_sents:
        gold_sentences = _read_sentence_scores(pipeline_options.gold_sents)
        golds[const.SENTENCE_SCORES] = gold_sentences
    return golds


def retrieve_predictions(pipeline_options, is_wmt18_pred_format):
    pred_files = {target: [] for target in const.TARGETS}
    if pipeline_options.pred_target:
        for pred_file in pipeline_options.pred_target:
            pred_target = read_file(pred_file)
            if is_wmt18_pred_format:
                pred_target, pred_gaps = _split_wmt18(pred_target)
                pred_files[const.GAP_TAGS].append((str(pred_file), pred_gaps))
            pred_files[const.TARGET_TAGS].append((str(pred_file), pred_target))
    if pipeline_options.pred_gaps:
        for pred_file in pipeline_options.pred_gaps:
            pred_gaps = read_file(pred_file)
            pred_files[const.GAP_TAGS].append((str(pred_file), pred_gaps))
    if pipeline_options.pred_source:
        for pred_file in pipeline_options.pred_source:
            pred_source = read_file(pred_file)
            pred_files[const.SOURCE_TAGS].append((str(pred_file), pred_source))
    if pipeline_options.pred_sents:
        for pred_file in pipeline_options.pred_sents:
            pred_sents = _read_sentence_scores(pred_file)
            pred_files[const.SENTENCE_SCORES].append(
                (str(pred_file), pred_sents)
            )
    if pipeline_options.input_dir:
        for input_dir in pipeline_options.input_dir:
            input_dir = Path(input_dir)
            for target in const.TAGS:
                pred_file = input_dir.joinpath(target)
                if pred_file.exists() and pred_file.is_file():
                    pred_files[pred_file.name].append(
                        (str(pred_file), read_file(pred_file))
                    )
            for target in [const.SENTENCE_SCORES, const.BINARY]:
                pred_file = input_dir.joinpath(target)
                if pred_file.exists() and pred_file.is_file():
                    pred_files[pred_file.name].append(
                        (str(pred_file), _read_sentence_scores(str(pred_file)))
                    )

    # Numericalize Text Labels
    if pipeline_options.type == "tags":
        for tag_name in const.TAGS:
            for i in range(len(pred_files[tag_name])):
                fname, pred_tags = pred_files[tag_name][i]
                pred_files[tag_name][i] = (fname, _wmt_to_labels(pred_tags))
    return pred_files


def retrieve_sentence_predictions(pipeline_options, pred_files):
    sent_preds = pred_files[const.SENTENCE_SCORES]
    sents_avg = (
        pipeline_options.sents_avg
        if pipeline_options.sents_avg
        else pipeline_options.type
    )
    tag_to_sent = _probs_to_sentence_score
    if sents_avg == "tags":
        tag_to_sent = _tags_to_sentence_score
    for pred_file, pred in pred_files[const.TARGET_TAGS]:
        sent_pred = np.array(tag_to_sent(pred))
        sent_preds.append((pred_file, sent_pred))
    return sent_preds


def _split_wmt18(tags):
    """Split tags list of lists in WMT18 format into target and gap tags."""
    tags_mt = [sent_tags[1::2] for sent_tags in tags]
    tags_gaps = [sent_tags[::2] for sent_tags in tags]
    return tags_mt, tags_gaps


def _wmt_to_labels(corpus):
    """Generates numeric labels from text labels."""
    dictionary = dict(zip(const.LABELS, range(len(const.LABELS))))
    return [[dictionary[word] for word in sent] for sent in corpus]


def _read_sentence_scores(sent_file):
    """Read File with numeric scores for sentences."""
    return np.array([line.strip() for line in open(sent_file)], dtype="float")


def _tags_to_sentence_score(tags_sentences):
    scores = []
    bad_label = const.LABELS.index(const.BAD)
    for tags in tags_sentences:
        labels = _probs_to_labels(tags)
        scores.append(labels.count(bad_label) / len(tags))
    return scores


def _probs_to_sentence_score(probs_sentences):
    scores = []
    for probs in probs_sentences:
        probs = [float(p) for p in probs]
        scores.append(np.mean(probs))
    return scores


def _probs_to_labels(probs, threshold=0.5):
    """Generates numeric labels from probabilities.

    This assumes two classes and default decision threshold 0.5

    """
    return [int(float(prob) > threshold) for prob in probs]


def _check_lengths(gold, prediction):
    for i, (g, p) in enumerate(zip(gold, prediction)):
        if len(g) != len(p):
            warnings.warn(
                "Mismatch length for {}th sample "
                "{} x {}".format(i, len(g), len(p))
            )


def _average(probs_per_file):
    # flat_probs = [list(flatten(probs)) for probs in probs_per_file]
    probabilities = np.array(probs_per_file, dtype="float32")
    return probabilities.mean(axis=0).tolist()


def _extract_path_prefix(file_names):
    if len(file_names) < 2:
        return "", file_names
    prefix_path = os.path.commonpath(
        [path for path in file_names if not path.startswith("*")]
    )
    if len(prefix_path) > 0:
        file_names = [
            os.path.relpath(path, prefix_path)
            if not path.startswith("*")
            else path
            for path in file_names
        ]
    return prefix_path, file_names


def setup():
    pass


def teardown():
    pass


def eval_word_level(golds, pred_files, tag_name):
    scores_table = []
    for pred_file, pred in pred_files[tag_name]:
        _check_lengths(golds[tag_name], pred)

        scores = score_word_level(
            list(flatten(golds[tag_name])), list(flatten(pred))
        )

        scores_table.append((pred_file, *scores))
    # If more than one system is provided, compute ensemble score
    if len(pred_files[tag_name]) > 1:
        ensemble_pred = _average(
            [list(flatten(pred)) for _, pred in pred_files[tag_name]]
        )
        ensemble_score = score_word_level(
            list(flatten(golds[tag_name])), ensemble_pred
        )
        scores_table.append(("*ensemble*", *ensemble_score))

    scores = np.array(
        scores_table,
        dtype=[
            ("File", "object"),
            ("F1_{}".format(const.LABELS[0]), float),
            ("F1_{}".format(const.LABELS[1]), float),
            ("F1_mult", float),
        ],
    )
    # Put the main metric in the first column
    scores = scores[
        [
            "File",
            "F1_mult",
            "F1_{}".format(const.LABELS[0]),
            "F1_{}".format(const.LABELS[1]),
        ]
    ]

    return scores


def eval_sentence_level(sent_gold, sent_preds):
    sentence_scores, sentence_ranking = [], []
    for file_name, pred in sent_preds:
        scoring, ranking = score_sentence_level(sent_gold, pred)

        sentence_scores.append((file_name, *scoring))
        sentence_ranking.append((file_name, *ranking))

    ensemble_pred = _average([pred for _, pred in sent_preds])
    ensemble_score, ensemble_ranking = score_sentence_level(
        sent_gold, ensemble_pred
    )
    sentence_scores.append(("*ensemble*", *ensemble_score))
    sentence_ranking.append(("*ensemble*", *ensemble_ranking))

    sentence_scores = np.array(
        sentence_scores,
        dtype=[
            ("File", "object"),
            ("Pearson r", float),
            ("MAE", float),
            ("RMSE", float),
        ],
    )
    sentence_ranking = np.array(
        sentence_ranking,
        dtype=[("File", "object"), ("Spearman r", float), ("DeltaAvg", float)],
    )

    return sentence_scores, sentence_ranking


def score_word_level(gold, prediction):
    gold_tags = gold
    pred_tags = _probs_to_labels(prediction)
    return f1_scores(pred_tags, gold_tags)


def score_sentence_level(gold, pred):
    pearson = pearsonr(gold, pred)
    mae = mean_absolute_error(gold, pred)
    rmse = np.sqrt(mean_squared_error(gold, pred))

    spearman = spearmanr(
        rankdata(gold, method="ordinal"), rankdata(pred, method="ordinal")
    )
    delta_avg = delta_average(gold, rankdata(pred, method="ordinal"))

    return (pearson[0], mae, rmse), (spearman[0], delta_avg)


def print_scores_table(scores, prefix="TARGET"):
    prefix_path, scores["File"] = _extract_path_prefix(scores["File"])
    path_str = " ({})".format(prefix_path) if prefix_path else ""

    max_method_length = max(len(path_str) + 4, max(map(len, scores["File"])))
    print("-" * (max_method_length + 13 * 3))
    print("Word-level scores for {}:".format(prefix))
    print(
        "{:{width}}    {:9}    {:9}    {:9}".format(
            "File{}".format(path_str),
            "F1_mult",
            "F1_{}".format(const.LABELS[0]),
            "F1_{}".format(const.LABELS[1]),
            width=max_method_length,
        )
    )
    for score in np.sort(scores, order=["F1_mult", "File"])[::-1]:
        print(
            "{:{width}s}    {:<9.5f}    {:<9.5}    {:<9.5f}".format(
                *score, width=max_method_length
            )
        )


def print_sentences_scoring_table(scores):
    prefix_path, scores["File"] = _extract_path_prefix(scores["File"])
    path_str = " ({})".format(prefix_path) if prefix_path else ""

    max_method_length = max(len(path_str) + 4, max(map(len, scores["File"])))
    print("-" * (max_method_length + 13 * 3))
    print("Sentence-level scoring:")
    print(
        "{:{width}}    {:9}    {:9}    {:9}".format(
            "File{}".format(path_str),
            "Pearson r",
            "MAE",
            "RMSE",
            width=max_method_length,
        )
    )
    for score in np.sort(scores, order=["Pearson r", "File"])[::-1]:
        print(
            "{:{width}s}    {:<9.5f}    {:<9.5f}    {:<9.5f}".format(
                *score, width=max_method_length
            )
        )


def print_sentences_ranking_table(scores):
    prefix_path, scores["File"] = _extract_path_prefix(scores["File"])
    path_str = " ({})".format(prefix_path) if prefix_path else ""

    max_method_length = max(len(path_str) + 4, max(map(len, scores["File"])))
    print("-" * (max_method_length + 13 * 3))
    print("Sentence-level ranking:")
    print(
        "{:{width}}    {:10}    {:9}".format(
            "File{}".format(path_str),
            "Spearman r",
            "DeltaAvg",
            width=max_method_length,
        )
    )  # noqa
    for score in np.sort(scores, order=["Spearman r", "File"])[::-1]:
        print(
            "{:{width}s}    {:<10.5f}    {:<9.5f}".format(
                *score, width=max_method_length
            )
        )



#  OpenKiwi: Open-Source Machine Translation Quality Estimation
#  Copyright (C) 2019 Unbabel <openkiwi@unbabel.com>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import logging
from pathlib import Path
from pprint import pformat

from kiwi.data.builders import build_test_dataset
from kiwi.data.utils import (
    deserialize_fields_from_vocabs,
    save_predicted_probabilities,
)
from kiwi.lib.utils import (
    configure_device,
    configure_logging,
    configure_seed,
    save_config_file,
    setup_output_directory,
)
from kiwi.models.linear_word_qe_classifier import LinearWordQEClassifier
from kiwi.models.model import Model
from kiwi.predictors.linear_tester import LinearTester
from kiwi.predictors.predictor import Predicter

logger = logging.getLogger(__name__)


def predict_from_options(options):
    """
    Uses the configuration options to run the prediction pipeline.
    Iteratively calls `setup`, `run` and `teardown`.

    Args:
        options (Namespace): Namespace containing all parsed options.
    """
    logger.debug("Setting up predict..")
    output_dir = setup(options.pipeline)

    logger.debug("Predict set up. Running...")
    run(options.model_api, output_dir, options.pipeline, options.model)

    logger.debug("Prediction finished. Tearing down")
    teardown(options.pipeline)


def load_model(model_path):
    """Load a pretrained model into a `Predicter` object.

    Args:
      load_model (str): A path to the saved model file.

    Throws:
      Exception: If the path does not exist, or is not a valid model file.

    """
    model_path = Path(model_path)
    if not model_path.exists():
        raise Exception('Path "{}" does not exist!'.format(model_path))

    model = Model.create_from_file(model_path)
    if not model:
        raise Exception('No model found in "{}"'.format(model_path))
    fieldset = model.fieldset()
    fields = deserialize_fields_from_vocabs(fieldset.fields, model.vocabs)
    predicter = Predicter(model, fields=fields)
    return predicter


def run(ModelClass, output_dir, pipeline_opts, model_opts):
    """
    Runs the prediction pipeline. Loads the model and necessary files
    and creates the model's predictions for all data received.

    Args:
        ModelClass (type): Python Type of the Model to train
        output_dir: Directory to save predictions
        pipeline_options (Namespace): Generic predict Options
            batch_size: Max batch size for predicting

        model_options (Namespace): Model Specific options

    Returns:
        Predictions (dict): Dictionary with format {'target':predictions}
    """
    model_name = getattr(ModelClass, "title", ModelClass.__name__)
    logger.info("Predict with the {} model".format(model_name))

    if ModelClass == LinearWordQEClassifier:
        load_vocab = None

        model = LinearWordQEClassifier(
            evaluation_metric=model_opts.evaluation_metric
        )
        model.load(pipeline_opts.load_model)
        predicter = LinearTester(model)
    else:
        load_vocab = pipeline_opts.load_model

        model = Model.create_from_file(pipeline_opts.load_model)

        # Set GPU or CPU. This has to be done before instantiating the optimizer
        device_id = None
        if pipeline_opts.gpu_id is not None and pipeline_opts.gpu_id >= 0:
            device_id = pipeline_opts.gpu_id
        model.to(device_id)

        predicter = Predicter(model)

    test_dataset = build_test_dataset(
        fieldset=ModelClass.fieldset(
            wmt18_format=model_opts.__dict__.get("wmt18_format")
        ),
        load_vocab=load_vocab,
        **vars(model_opts),
    )
    predictions = predicter.run(
        test_dataset, batch_size=pipeline_opts.batch_size
    )

    save_predicted_probabilities(output_dir, predictions)
    return predictions


def setup(options):
    """
    Analyze pipeline options and set up requirements to running
    the prediction pipeline. This includes setting up the output
    directory, random seeds and the device where predictions are run.

    Args:
        options(Namespace): Pipeline specific options

    Returns:
        output_dir(str): Path to output directory
    """
    output_dir = setup_output_directory(
        options.output_dir, options.run_uuid, experiment_id=None, create=True
    )
    configure_logging(
        output_dir=output_dir, debug=options.debug, quiet=options.quiet
    )
    configure_seed(options.seed)
    configure_device(options.gpu_id)

    logger.info(pformat(vars(options)))
    logger.info("Local output directory is: {}".format(output_dir))

    if options.save_config:
        save_config_file(options, options.save_config)

    del options.output_dir  # FIXME: remove this after making sure no other
    # place uses it! # noqa
    return output_dir


def teardown(options):
    """
    Tears down after executing prediction pipeline.

    Args:
        options(Namespace): Pipeline specific options
    """
    pass



#  OpenKiwi: Open-Source Machine Translation Quality Estimation
#  Copyright (C) 2019 Unbabel <openkiwi@unbabel.com>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import logging
from collections import defaultdict
from pathlib import Path

import torch
from tqdm import tqdm

import kiwi
from kiwi import constants as const
from kiwi.loggers import tracking_logger
from kiwi.metrics.stats import Stats
from kiwi.models.model import Model
from kiwi.models.utils import load_torch_file
from kiwi.trainers.callbacks import EarlyStopException
from kiwi.trainers.utils import optimizer_class

logger = logging.getLogger(__name__)


class Trainer:
    def __init__(
        self, model, optimizer, checkpointer, log_interval=100, scheduler=None
    ):
        """
        Args:
          model: A kiwi.Model to train
          optimizer: An optimizer
          checkpointer: A Checkpointer object
          log_interval: Log train stats every /n/ batches. Default 100
          scheduler: A learning rate scheduler
        """
        self.model = model
        self.stats = Stats(
            metrics=model.metrics(),
            main_metric_ordering=model.metrics_ordering(),
            log_interval=log_interval,
        )

        self.optimizer = optimizer
        self.checkpointer = checkpointer
        self.scheduler = scheduler
        self._step = 0
        self._epoch = 0

    @property
    def stats_summary_history(self):
        return self.checkpointer.stats_summary_history

    def run(self, train_iterator, valid_iterator, epochs=50):
        """

        Args:
            train_iterator:
            epochs: Number of epochs for training.
        """
        # log(self.eval_epoch(valid_dataset))
        for epoch in range(self._epoch + 1, epochs + 1):
            logger.info('Epoch {} of {}'.format(epoch, epochs))
            self.train_epoch(train_iterator, valid_iterator)
            self.stats.log()
            try:
                self.checkpointer(self, valid_iterator, epoch=epoch)
            except EarlyStopException as e:
                logger.info(e)
                break

        self.checkpointer.check_out()

    def train_epoch(self, train_iterator, valid_iterator):
        self.model.train()
        for batch in tqdm(
            train_iterator,
            total=len(train_iterator),
            desc='Batches',
            unit=' batches',
            ncols=80,
        ):
            self._step += 1
            outputs = self.train_step(batch)
            self.stats.update(batch=batch, **outputs)
            self.stats.log(step=self._step)
            try:
                self.checkpointer(self, valid_iterator, step=self._step)
            except EarlyStopException as e:
                logger.info(e)
                break
        self._epoch += 1

    def train_steps(self, train_iterator, valid_iterator, max_steps):
        train_iterator.repeat = True
        self.model.train()
        step = 0
        for step, batch in tqdm(
            enumerate(train_iterator, 1),
            total=max_steps,
            desc='Steps',
            unit=' batches',
            ncols=80,
        ):
            self._step += 1
            outputs = self.train_step(batch)
            self.stats.update(batch=batch, **outputs)
            self.stats.log(step=self._step)
            try:
                self.checkpointer(self, valid_iterator, step=self._step)
            except EarlyStopException as e:
                logger.info(e)
                break

            if step > max_steps:
                break

        eval_stats_summary = self.eval_epoch(valid_iterator)
        eval_stats_summary.log()

        sub_path = Path('step_{}'.format(self._step))
        self.save(self.checkpointer.output_directory / sub_path)

        train_iterator.repeat = False

    def train_step(self, batch):
        self.model.zero_grad()
        model_out = self.model(batch)
        loss_dict = self.model.loss(model_out, batch)
        loss_dict[const.LOSS].backward()
        self.optimizer.step()
        return dict(loss=loss_dict, model_out=model_out)

    def eval_epoch(self, valid_iterator, prefix='EVAL'):
        self.model.eval()
        self.stats.reset()
        with torch.no_grad():
            for batch in valid_iterator:
                outputs = self.eval_step(batch)
                self.stats.update(batch=batch, **outputs)
        stats_summary = self.stats.wrap_up(prefix=prefix)
        self.model.train()
        return stats_summary

    def eval_step(self, batch):
        model_out = self.model(batch)
        loss_dict = self.model.loss(model_out, batch)
        return dict(loss=loss_dict, model_out=model_out)

    def predict(self, valid_iterator):
        self.model.eval()
        with torch.no_grad():
            predictions = defaultdict(list)
            for batch in valid_iterator:
                model_pred = self.model.predict(batch)
                for key, values in model_pred.items():
                    predictions[key] += values
        self.model.train()
        return predictions

    def make_sub_directory(self, root_directory, current_epoch, prefix='epoch'):
        root_path = Path(root_directory)
        epoch_path = Path('{}_{}'.format(prefix, current_epoch))
        output_path = root_path / epoch_path
        output_path.mkdir(exist_ok=True)
        return output_path

    def save(self, output_directory):
        output_directory = Path(output_directory)
        output_directory.mkdir(exist_ok=True)
        logging.info('Saving training state to {}'.format(output_directory))

        model_path = output_directory / const.MODEL_FILE
        self.model.save(str(model_path))

        optimizer_path = output_directory / const.OPTIMIZER
        scheduler_dict = None
        if self.scheduler:
            scheduler_dict = {
                'name': type(self.scheduler).__name__.lower(),
                'state_dict': self.scheduler.state_dict(),
            }
        optimizer_dict = {
            'name': type(self.optimizer).__name__.lower(),
            'state_dict': self.optimizer.state_dict(),
            'scheduler_dict': scheduler_dict,
        }
        torch.save(optimizer_dict, str(optimizer_path))

        state = {
            '__version__': kiwi.__version__,
            '_epoch': self._epoch,
            '_step': self._step,
            'checkpointer': self.checkpointer,
        }
        state_path = output_directory / const.TRAINER
        torch.save(state, str(state_path))

        # Send to MLflow
        event = None
        if tracking_logger.should_log_artifacts():
            logger.info('Logging artifacts to {}'.format(output_directory))
            event = tracking_logger.log_artifacts(
                str(output_directory), artifact_path=str(output_directory.name)
            )
        return event

    @classmethod
    def from_directory(cls, directory, device_id=None):
        logger.info('Loading training state from {}'.format(directory))
        root_path = Path(directory)

        model_path = root_path / const.MODEL_FILE
        model = Model.create_from_file(model_path)

        if device_id is not None:
            model.to(device_id)

        optimizer_path = root_path / const.OPTIMIZER
        optimizer_dict = load_torch_file(str(optimizer_path))

        optimizer = optimizer_class(optimizer_dict['name'])(
            model.parameters(), lr=0.0
        )
        optimizer.load_state_dict(optimizer_dict['state_dict'])

        trainer = cls(model, optimizer, checkpointer=None)
        trainer_path = root_path / const.TRAINER
        state = load_torch_file(str(trainer_path))
        trainer.__dict__.update(state)
        return trainer

    @classmethod
    def resume(cls, local_path=None, prefix='latest_', device_id=None):
        if local_path:
            artifacts_uri = Path(local_path)
        else:
            artifacts_uri = Path(tracking_logger.get_artifact_uri())

        if Path(local_path) / Path(prefix + 'epoch') in artifacts_uri.glob(
            '{}*'.format(prefix)
        ):
            last_save = 'epoch'

        else:
            logging.info(
                'Latest epoch not found. Looking for other checkpoints'
            )
            prefix = 'epoch_'
            saved_checkpoints = [
                int(str(path.name).replace(prefix, ''))
                for path in artifacts_uri.glob('{}*'.format(prefix))
                if path.is_dir()
            ]
            if not saved_checkpoints:
                raise FileNotFoundError(
                    "Couldn't load trainer from: {}".format(
                        artifacts_uri / (prefix + '*')
                    )
                )
            last_save = max(saved_checkpoints)

        snapshot_dir = artifacts_uri / '{}{}'.format(prefix, last_save)
        logger.info('Resuming training from: {}'.format(snapshot_dir))
        return cls.from_directory(snapshot_dir, device_id=device_id)



#  OpenKiwi: Open-Source Machine Translation Quality Estimation
#  Copyright (C) 2019 Unbabel <openkiwi@unbabel.com>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

from torchtext import data


class AlignmentField(data.Field):
    def process(self, batch, *args, **kwargs):
        """ Process a list of examples to create a batch.

        Postprocess the batch with user-provided Pipeline.

        Args:
            batch (list(object)): A list of object from a batch of examples.
        Returns:
            object: Processed object given the input and custom
                postprocessing Pipeline.
        """
        if self.postprocessing is not None:
            batch = self.postprocessing(batch)
        return batch



#  OpenKiwi: Open-Source Machine Translation Quality Estimation
#  Copyright (C) 2019 Unbabel <openkiwi@unbabel.com>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

from torchtext import data

from kiwi import constants as const
from kiwi.data import utils
from kiwi.data.fields.alignment_field import AlignmentField
from kiwi.data.fields.qe_field import QEField
from kiwi.data.fields.sequence_labels_field import SequenceLabelsField
from kiwi.data.fieldsets.fieldset import Fieldset
from kiwi.data.tokenizers import align_tokenizer, tokenizer


def build_fieldset(wmt18_format=False):
    fs = Fieldset()

    fs.add(
        name=const.SOURCE,
        field=QEField(
            tokenize=tokenizer,
            init_token=None,
            eos_token=None,
            include_lengths=True,
        ),
        file_option_suffix='_source',
        required=Fieldset.ALL,
        vocab_options=dict(
            min_freq='source_vocab_min_frequency',
            max_size='source_vocab_size',
            rare_with_vectors='keep_rare_words_with_embeddings',
            add_vectors_vocab='add_embeddings_vocab',
        ),
        vocab_vectors='source_embeddings',
    )
    fs.add(
        name=const.TARGET,
        field=QEField(
            tokenize=tokenizer,
            init_token=None,
            eos_token=None,
            include_lengths=True,
        ),
        file_option_suffix='_target',
        required=Fieldset.ALL,
        vocab_options=dict(
            min_freq='target_vocab_min_frequency',
            max_size='target_vocab_size',
            rare_with_vectors='keep_rare_words_with_embeddings',
            add_vectors_vocab='add_embeddings_vocab',
        ),
        vocab_vectors='target_embeddings',
    )

    fs.add(
        name=const.ALIGNMENTS,
        field=AlignmentField(tokenize=align_tokenizer, use_vocab=False),
        file_option_suffix='_alignments',
        required=Fieldset.ALL,
    )

    post_pipe_target = data.Pipeline(utils.project)
    if wmt18_format:
        post_pipe_gaps = data.Pipeline(utils.wmt18_to_gaps)
        post_pipe_target = data.Pipeline(utils.wmt18_to_target)
        fs.add(
            name=const.GAP_TAGS,
            field=SequenceLabelsField(
                classes=const.LABELS,
                tokenize=tokenizer,
                pad_token=const.PAD,
                unk_token=None,
                batch_first=True,
                # eos_token=const.STOP,
                postprocessing=post_pipe_gaps,
            ),
            file_option_suffix='_target_tags',
            required=[Fieldset.TRAIN, Fieldset.VALID],
        )

    fs.add(
        name=const.TARGET_TAGS,
        field=SequenceLabelsField(
            classes=const.LABELS,
            tokenize=tokenizer,
            pad_token=const.PAD,
            unk_token=None,
            batch_first=True,
            postprocessing=post_pipe_target,
        ),
        file_option_suffix='_target_tags',
        required=[Fieldset.TRAIN, Fieldset.VALID],
    )
    fs.add(
        name=const.SOURCE_TAGS,
        field=SequenceLabelsField(
            classes=const.LABELS,
            tokenize=tokenizer,
            pad_token=const.PAD,
            unk_token=None,
            batch_first=True,
        ),
        file_option_suffix='_source_tags',
        required=None,
    )

    return fs



#  OpenKiwi: Open-Source Machine Translation Quality Estimation
#  Copyright (C) 2019 Unbabel <openkiwi@unbabel.com>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import logging
from pathlib import Path

import numpy as np
import torch
import torch.nn.functional as F
from more_itertools import first, flatten
from torch.autograd import Function
from torch.nn.utils.rnn import pack_padded_sequence as pack
from torch.nn.utils.rnn import pad_packed_sequence as unpack

from kiwi import constants as const

logger = logging.getLogger(__name__)


def unroll(list_of_lists):
    """
    :param list_of_lists: a list that contains lists
    :param rec: unroll recursively
    :return: a flattened list
    """
    if isinstance(first(list_of_lists), (np.ndarray, list)):
        return list(flatten(list_of_lists))
    return list_of_lists


def convolve_tensor(sequences, window_size, pad_value=0):
    """Convolve a sequence and apply padding

    :param sequence: 2D tensor
    :param window_size: filter length
    :param pad_value: int value used as padding
    :return: 3D tensor, where the last dimension has size window_size
    """
    pad = (window_size // 2,) * 2
    t = F.pad(sequences, pad=pad, value=pad_value)
    t = t.unfold(1, window_size, 1)

    # For 3D tensors
    # torch.nn.ConstantPad2d((0, 0, 1, 1), 0)(x).unfold(1, 3, 1)
    # F.pad(x, (0, 0, 1, 1), value=0).unfold(1, 3, 1)

    return t


# def convolve_sequence(sequence, window_size, pad_value=0):
#     """Convolve a sequence and apply padding
#
#     :param sequence: list of ids
#     :param window_size: filter length
#     :param pad_value: int value used as padding
#     :return: list of lists with size of window_size
#     """
#     pad = [pad_value for _ in range(window_size // 2)]
#     pad_sequence = pad + sequence + pad
#     return list(windowed(pad_sequence, window_size, fillvalue=pad_value))


def align_tensor(
    tensor,
    alignments,
    max_aligned,
    unaligned_idx,
    padding_idx,
    pad_size,
    target_length=None,
):
    alignments = [
        map_alignments_to_target(sample, target_length=target_length)
        for sample in alignments
    ]

    # aligned_tensor = tensor.new_full(
    #     (tensor.shape[0], pad_size, max_aligned, tensor.shape[2]),
    #     padding_idx)

    aligned = [
        align_source(
            sample, alignment, max_aligned, unaligned_idx, padding_idx, pad_size
        )
        for sample, alignment in zip(tensor, alignments)
    ]
    aligned_tensor = torch.stack([sample[0] for sample in aligned])
    nb_alignments = torch.stack([sample[1] for sample in aligned])
    return aligned_tensor, nb_alignments


def map_alignments_to_target(src2tgt_alignments, target_length=None):
    """Maps a target index to a list of source indexes.

    Args:
        src2tgt_alignments (list): list of tuples with source, target indexes.
        target_length: size of the target side; if None, the highest index
            in the alignments is used.

    Returns:
        A list of size target_length where position i refers to the i-th
        target token and contains a list of source indexes aligned to it.

    """
    if target_length is None:
        if not src2tgt_alignments:
            target_length = 0
        else:
            target_length = 1 + max(src2tgt_alignments, key=lambda a: a[1])[1]

    trg2src = [None] * target_length
    for source, target in src2tgt_alignments:
        if not trg2src[target]:
            trg2src[target] = []
        trg2src[target].append(source)
    return trg2src


def align_source(
    source,
    trg2src_alignments,
    max_aligned,
    unaligned_idx,
    padding_idx,
    pad_size,
):
    assert len(source.shape) == 2
    window_size = source.shape[1]

    assert len(trg2src_alignments) <= pad_size
    aligned_source = source.new_full(
        (pad_size, max_aligned, window_size), padding_idx
    )
    unaligned = source.new_full((window_size,), unaligned_idx)
    nb_alignments = source.new_ones(pad_size, dtype=torch.float)

    for i, source_positions in enumerate(trg2src_alignments):
        if not source_positions:
            aligned_source[i, 0] = unaligned
        else:
            selected = torch.index_select(
                source,
                0,
                torch.tensor(
                    source_positions[:max_aligned], device=source.device
                ),
            )
            aligned_source[i, : len(selected)] = selected
            # counts how many tokens is a target token aligned to
            nb_alignments[i] = len(selected)
    return aligned_source, nb_alignments


def apply_packed_sequence(rnn, embedding, lengths):
    """ Runs a forward pass of embeddings through an rnn using packed sequence.
    Args:
       rnn: The RNN that that we want to compute a forward pass with.
       embedding (FloatTensor b x seq x dim): A batch of sequence embeddings.
       lengths (LongTensor batch): The length of each sequence in the batch.

    Returns:
       output: The output of the RNN `rnn` with input `embedding`
    """
    # Sort Batch by sequence length
    lengths_sorted, permutation = torch.sort(lengths, descending=True)
    embedding_sorted = embedding[permutation]

    # Use Packed Sequence
    embedding_packed = pack(embedding_sorted, lengths_sorted, batch_first=True)
    outputs_packed, (hidden, cell) = rnn(embedding_packed)
    outputs_sorted, _ = unpack(outputs_packed, batch_first=True)
    # Restore original order
    _, permutation_rev = torch.sort(permutation, descending=False)
    outputs = outputs_sorted[permutation_rev]
    hidden, cell = hidden[:, permutation_rev], cell[:, permutation_rev]
    return outputs, (hidden, cell)


def replace_token(target, old, new):
    """Replaces old tokens with new.

    args: target (LongTensor)
          old (int): The token to be replaced by new
          new (int): The token used to replace old

    """
    return target.masked_fill(target == old, new)


def make_loss_weights(nb_classes, target_idx, weight):
    """Creates a loss weight vector for nn.CrossEntropyLoss

    args:
        nb_classes: Number of classes
        target_idx: ID of the target (reweighted) class
        weight: Weight of the target class

    returns:
       weights (FloatTensor): Weight Tensor of shape `nb_classes` such that
                                  `weights[target_idx] = weight`
                                  `weights[other_idx] = 1.0`
    """

    weights = torch.ones(nb_classes)
    weights[target_idx] = weight
    return weights


def load_torch_file(file_path):
    file_path = Path(file_path)
    if not file_path.exists():
        raise FileNotFoundError('Torch file not found: {}'.format(file_path))

    file_dict = torch.load(
        str(file_path), map_location=lambda storage, loc: storage
    )
    if isinstance(file_dict, Path):
        # Resolve cases where file is just a link to another torch file
        linked_path = file_dict
        if not linked_path.exists():
            relative_path = (
                file_path.with_name(file_dict.name) / const.MODEL_FILE
            )
            if relative_path.exists():
                linked_path = relative_path
        return load_torch_file(linked_path)
    return file_dict


class GradientMul(Function):
    @staticmethod
    def forward(ctx, x, constant=0):
        ctx.constant = constant
        return x

    @staticmethod
    def backward(ctx, grad):
        return ctx.constant * grad, None


gradient_mul = GradientMul.apply



#  OpenKiwi: Open-Source Machine Translation Quality Estimation
#  Copyright (C) 2019 Unbabel <openkiwi@unbabel.com>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import argparse
import logging
import random
from argparse import Namespace
from pathlib import Path
from time import gmtime

import configargparse
import numpy as np
import torch


def configure_seed(seed):
    """
    Configure the random seed for all relevant packages.
    These include: random, numpy, torch and torch.cuda

    Args:
        seed (int): the random seed to be set
    """
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)


def configure_device(gpu_id):
    """
    Configure gpu to be used in computation.

    Args:
        gpu_id (int): The id of the gpu to be used
    """
    if gpu_id is not None:
        torch.cuda.set_device(gpu_id)


def configure_logging(output_dir=None, debug=False, quiet=False):
    """
    Configure the logger. Sets up the log format, logging level
    and output directory of logging.

    Args:
        output_dir: The directory where log output will be stored.
            Defaults to None.
        debug (bool): Change logging level to debug.
        quiet (bool): Change logging level to warning to supress info logs.
    """
    logging.Formatter.converter = gmtime
    logging.Formatter.default_msec_format = '%s.%03d'
    log_format = '%(asctime)s [%(name)s %(funcName)s:%(lineno)s] %(message)s'
    if logging.getLogger().handlers:
        log_formatter = logging.Formatter(log_format)
        for handler in logging.getLogger().handlers:
            handler.setFormatter(log_formatter)
    else:
        logging.basicConfig(level=logging.INFO, format=log_format)

    log_level = logging.INFO
    if debug:
        log_level = logging.DEBUG
    if quiet:
        log_level = logging.WARNING

    logging.getLogger().setLevel(log_level)
    if output_dir is not None:
        fh = logging.FileHandler(str(Path(output_dir, 'output.log')))
        fh.setLevel(log_level)
        logging.getLogger().addHandler(fh)


def save_args_to_file(file_name, **kwargs):
    """
    Saves `**kwargs` to a file.

    Args:
        file_name (str): The name of the file where the args should
            be saved in.

    """
    options_to_save = {
        k.replace('_', '-'): v for k, v in kwargs.items() if v is not None
    }

    content = configargparse.YAMLConfigFileParser().serialize(options_to_save)
    Path(file_name).write_text(content)
    logging.debug('Saved current options to config file: {}'.format(file_name))


def save_config_file(options, file_name):
    """
    Saves a configuration file with OpenKiwi configuration options.
    Calls `save_args_to_file`.

    Args:
        options (Namespace): Namespace with all configuration options
            that should be saved.
        file_name (str): Name of the output configuration file.
    """
    # parser.write_config_file(options, [file_name], exit_after=False)
    save_args_to_file(file_name, **vars(options))


def setup_output_directory(
    output_dir, run_uuid=None, experiment_id=None, create=True
):
    """
    Sets up the output directory. This means either creating one, or
    verifying that the provided directory exists. Output directories
    are created using the run and experiment ids.

    Args:
        output_dir (str): The target output directory
        run_uuid : The current hash of the current run.
        experiment_id: The id of the current experiment
        create (bool): Boolean indicating whether to create a new folder.
    """
    if not output_dir:
        if experiment_id is None or run_uuid is None:
            raise argparse.ArgumentError(
                message='Please specify an output directory (--output-dir).',
                argument=output_dir,
            )
        output_path = Path('runs', str(experiment_id), str(run_uuid))
        output_dir = str(output_path)

    if create:
        Path(output_dir).mkdir(parents=True, exist_ok=True)
    elif not Path(output_dir).exists():
        raise FileNotFoundError(
            'Output directory does not exist: {}'.format(output_dir)
        )

    return output_dir


def merge_namespaces(*args):
    """
    Utility function used to merge Namespaces. Useful for merging Argparse
    options.

    Args:
        *args: Variable length list of Namespaces
    """
    if not args:
        return None
    options = {}
    for arg in filter(None, args):
        options.update(dict(vars(arg)))
    return Namespace(**options)


def parse_integer_with_positive_infinity(string):
    """
    Workaround to be able to pass both integers and infinity as CLAs.

    Args:
        string: A string representation of an integer, or infinity

    """
    try:
        integer = int(string)
        return integer
    except ValueError:
        infinity = float(string)
        if infinity == float('inf'):
            return infinity

    raise ValueError(
        'Could not parse argument "{}" as integer'
        ' with positive infinity'.format(string)
    )



# -*- coding: utf-8 -*-
"""This implements a dictionary of labels."""

#  OpenKiwi: Open-Source Machine Translation Quality Estimation
#  Copyright (C) 2019 Unbabel <openkiwi@unbabel.com>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import warnings


class LabelDictionary(dict):
    """This class implements a dictionary of labels. Labels as mapped to
    integers, and it is efficient to retrieve the label name from its
    integer representation, and vice-versa."""

    def __init__(self, label_names=None):
        dict.__init__(self)
        self.names = []
        if label_names is not None:
            for name in label_names:
                self.add(name)

    def add(self, name):
        """Add new label."""
        label_id = len(self.names)
        if name in self:
            warnings.warn('Ignoring duplicated label ' + name)
        self[name] = label_id
        self.names.append(name)
        return label_id

    def get_label_name(self, label_id):
        """Get label name from id."""
        return self.names[label_id]

    def get_label_id(self, name):
        """Get label id from name."""
        return self[name]

    def save(self, label_file):
        """Save labels to a file."""
        f = open(label_file, 'w')
        for name in self.names:
            f.write(name + '\n')
        f.close()

    def load(self, label_file):
        """Load labels from a file."""
        self.names = []
        self.clear()
        f = open(label_file)
        for line in f:
            name = line.rstrip('\n')
            self.add(name)
        f.close()



"""A class for handling features for word-level quality estimation."""

#  OpenKiwi: Open-Source Machine Translation Quality Estimation
#  Copyright (C) 2019 Unbabel <openkiwi@unbabel.com>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import numpy as np

from kiwi.models.linear.linear_word_qe_sentence import LinearWordQESentence

from .sparse_feature_vector import SparseFeatureVector


def quantize(value, bins_down):
    """Quantize a numeric feature into bins.
    Example: bins = [50, 40, 30, 25, 20, 18, 16, 14, 12, 10]."""
    bin_up = np.inf
    for bin_down in bins_down:
        if bin_down < value <= bin_up:
            bin_value = bin_down
            return bin_value
        bin_up = bin_down
    return value


class LinearWordQEFeatures(SparseFeatureVector):
    """This class implements a feature vector for word-level quality
    estimation."""

    def __init__(
        self,
        use_basic_features_only=True,
        use_simple_bigram_features=True,
        use_parse_features=False,
        use_stacked_features=False,
        save_to_cache=False,
        load_from_cache=False,
        cached_features_file=None,
    ):
        SparseFeatureVector.__init__(
            self, save_to_cache, load_from_cache, cached_features_file
        )
        self.use_basic_features_only = use_basic_features_only
        # True for using only a single bigram indicator feature.
        self.use_simple_bigram_features = use_simple_bigram_features
        self.use_parse_features = use_parse_features
        self.use_stacked_features = use_stacked_features
        self.use_client_features = False

    def get_siblings(self, sentence_word_features, index):
        if index < 0 or index >= len(sentence_word_features):
            info = None
        else:
            info = sentence_word_features[index]
        if info is not None:
            siblings = [
                k
                for k in range(len(sentence_word_features))
                if sentence_word_features[k].target_head == info.target_head
            ]
            left_siblings = [k for k in siblings if k < index]
            right_siblings = [k for k in siblings if k > index]
            if len(left_siblings) > 0:
                left_sibling = max(left_siblings)
            else:
                left_sibling = -1
            if len(right_siblings) > 0:
                right_sibling = min(right_siblings)
            else:
                right_sibling = -1
        else:
            left_sibling = -2
            right_sibling = -2

        if left_sibling >= 0:
            left_sibling_info = sentence_word_features[left_sibling]
            left_sibling_token = left_sibling_info.token
            left_sibling_pos = left_sibling_info.target_pos
        elif left_sibling == -1:
            left_sibling_token = '__ROOT__'
            left_sibling_pos = '__ROOT__'
        else:
            left_sibling_info = None
            left_sibling_token = '__START__'
            left_sibling_pos = '__START__'
        if right_sibling >= 0:
            right_sibling_info = sentence_word_features[right_sibling]
            right_sibling_token = right_sibling_info.token
            right_sibling_pos = right_sibling_info.target_pos
        elif right_sibling == -1:
            right_sibling_info = None
            right_sibling_token = '__ROOT__'
            right_sibling_pos = '__ROOT__'
        else:
            right_sibling_info = None
            right_sibling_token = '__START__'
            right_sibling_pos = '__START__'

        return (
            left_sibling_token,
            left_sibling_pos,
            right_sibling_token,
            right_sibling_pos,
        )

    def get_head(self, sentence_word_features, index):
        if index < 0 or index >= len(sentence_word_features):
            info = None
        else:
            info = sentence_word_features[index]
        if info is not None:
            head_index = info.target_head - 1
        else:
            head_index = -2
        if head_index >= 0:
            head_info = sentence_word_features[head_index]
            head_token = head_info.token
            head_pos = head_info.target_pos
            head_morph = head_info.target_morph
        elif head_index == -1:
            head_info = None
            head_token = '__ROOT__'
            head_pos = '__ROOT__'
            head_morph = '__ROOT__'
        else:
            head_info = None
            head_token = '__START__'
            head_pos = '__START__'
            head_morph = '__START__'

        return head_index, head_token, head_pos, head_morph

    def compute_unigram_features(self, sentence_word_features, part):
        """Compute unigram features (depending only on a single label)."""
        if self.load_from_cache:
            self.load_cached_features()
            return

        index = part.index

        ignore_source = False
        only_basic_features = self.use_basic_features_only
        use_client_features = self.use_client_features
        use_parse_features = self.use_parse_features
        use_stacked_features = self.use_stacked_features
        use_bias = True
        use_language_model = True
        use_binary_features = False
        if use_parse_features:
            use_split_morphs = False
            use_morph_features = False
            use_deprel_features = True
            use_head_features = True
            use_grandparent_features = True
            use_sibling_features = True
        else:
            use_split_morphs = False
            use_morph_features = False
            use_deprel_features = False
            use_head_features = False
            use_grandparent_features = False
            use_sibling_features = False
        use_unuseful_shared_task_features = False

        info = sentence_word_features[index]
        if use_client_features:
            labels = [str(part.label), info.client_name + '_' + str(part.label)]
        else:
            labels = [str(part.label)]

        for label in labels:
            if use_bias:
                self.add_binary_feature('BIAS_%s' % label)

            if use_unuseful_shared_task_features:
                self.add_binary_feature(
                    'F0=%d_%s'
                    % (
                        quantize(info.source_token_count, [40, 30, 20, 10]),
                        label,
                    )
                )
                self.add_binary_feature(
                    'F1=%d_%s'
                    % (
                        quantize(info.target_token_count, [40, 30, 20, 10]),
                        label,
                    )
                )
                self.add_binary_feature(
                    'F2=%f_%s'
                    % (
                        quantize(
                            info.source_target_token_count_ratio, [5.0, 2.0]
                        ),
                        label,
                    )
                )

            self.add_binary_feature('F3=%s_%s' % (info.token, label))
            self.add_binary_feature('F4=%s_%s' % (info.left_context, label))
            self.add_binary_feature('F5=%s_%s' % (info.right_context, label))

            if not ignore_source:
                self.add_binary_feature(
                    'F6=%s_%s' % (info.first_aligned_token, label)
                )
                self.add_binary_feature(
                    'F7=%s_%s' % (info.left_alignment, label)
                )
                self.add_binary_feature(
                    'F8=%s_%s' % (info.right_alignment, label)
                )

            if use_binary_features and not only_basic_features:
                # Ablated for German WMT16 (the provided stoplist is wrong).
                # self.add_binary_feature(
                #     'F9=%d_%s' % (int(info.is_stopword), label))
                self.add_binary_feature(
                    'F10=%d_%s' % (int(info.is_punctuation), label)
                )
                # Ablated for German (capitalized words are nouns)
                # self.add_binary_feature(
                #     'F11=%d_%s' % (int(info.is_proper_noun), label))
                self.add_binary_feature(
                    'F12=%d_%s' % (int(info.is_digit), label)
                )

            if use_language_model and not only_basic_features:
                self.add_binary_feature(
                    'F13=%d_%s' % (info.highest_order_ngram_left, label)
                )
                self.add_binary_feature(
                    'F14=%d_%s' % (info.highest_order_ngram_right, label)
                )

            # if use_language_model and not only_basic_features:
            #     self.add_binary_feature(
            #         'F15=%d_%s' % (info.backoff_behavior_left, label))
            #     self.add_binary_feature(
            #         'F16=%d_%s' % (info.backoff_behavior_middle, label))
            #     self.add_binary_feature(
            #         'F17=%d_%s' % (info.backoff_behavior_right, label))

            if use_language_model and not only_basic_features:
                self.add_binary_feature(
                    'F18=%d_%s' % (info.source_highest_order_ngram_left, label)
                )
                self.add_binary_feature(
                    'F19=%d_%s' % (info.source_highest_order_ngram_right, label)
                )
                self.add_binary_feature(
                    'F20=%d_%s' % (int(info.pseudo_reference), label)
                )

            if not only_basic_features:
                self.add_binary_feature('F21=%s_%s' % (info.target_pos, label))
                self.add_binary_feature(
                    'F22=%s_%s' % (info.aligned_source_pos_list, label)
                )

            if use_unuseful_shared_task_features:
                self.add_binary_feature(
                    'F23=%d_%s' % (info.polysemy_count_source, label)
                )
                self.add_binary_feature(
                    'F24=%d_%s' % (info.polysemy_count_target, label)
                )

            # QUETCH linear model conjoined features.
            self.add_binary_feature(
                'G0=%s_%s_%s' % (info.token, info.left_context, label)
            )
            self.add_binary_feature(
                'G1=%s_%s_%s' % (info.token, info.right_context, label)
            )
            if not ignore_source:
                self.add_binary_feature(
                    'G2=%s_%s_%s'
                    % (info.token, info.first_aligned_token, label)
                )
            if not only_basic_features:
                self.add_binary_feature(
                    'G3=%s_%s_%s'
                    % (info.target_pos, info.aligned_source_pos_list, label)
                )

            # Parse features.
            if use_parse_features:
                head_index, head_token, head_pos, head_morph = self.get_head(
                    sentence_word_features, index
                )
                head_on_left = True  # (head_index <= index)

                if head_index >= 0:
                    _, grandparent_token, grandparent_pos, _ = self.get_head(
                        sentence_word_features, head_index
                    )
                else:
                    grandparent_token, grandparent_pos = head_token, head_pos
                grandparent_on_left = True  # (grandparent_index <= index)

                left_sibling_token, left_sibling_pos, right_sibling_token, right_sibling_pos = self.get_siblings(  # NOQA
                    sentence_word_features, index
                )

                if use_deprel_features:
                    self.add_binary_feature(
                        'H0=%s_%s' % (info.target_deprel, label)
                    )
                    self.add_binary_feature(
                        'H1=%s_%s_%s' % (info.token, info.target_deprel, label)
                    )

                if use_head_features:
                    # self.add_binary_feature(
                    #     'H2=%s_%s_%s' % (info.target_pos, head_pos, label))
                    # self.add_binary_feature(
                    #     'H3=%s_%s_%s' % (info.token, head_token, label))
                    self.add_binary_feature(
                        'H2=%s_%s_%d_%s'
                        % (info.target_pos, head_pos, int(head_on_left), label)
                    )
                    self.add_binary_feature(
                        'H3=%s_%s_%d_%s'
                        % (info.token, head_token, int(head_on_left), label)
                    )
                    self.add_binary_feature(
                        'H3a=%s_%s_%d_%s'
                        % (info.token, head_pos, int(head_on_left), label)
                    )
                    self.add_binary_feature(
                        'H3b=%s_%s_%d_%s'
                        % (
                            info.target_pos,
                            head_token,
                            int(head_on_left),
                            label,
                        )
                    )

                if use_morph_features:
                    self.add_binary_feature(
                        'H4=%s_%s' % (info.target_morph, label)
                    )
                    self.add_binary_feature(
                        'H5=%s_%s_%s' % (info.target_morph, head_morph, label)
                    )

                if use_split_morphs:
                    all_morphs = info.target_morph.split('|')
                    all_head_morphs = head_morph.split('|')
                    for m in all_morphs:
                        self.add_binary_feature('H6=%s_%s' % (m, label))
                        for hm in all_head_morphs:
                            self.add_binary_feature(
                                'H7=%s_%s_%s' % (m, hm, label)
                            )

                if use_sibling_features:
                    self.add_binary_feature(
                        'H8=%s_%s_%s'
                        % (info.target_pos, left_sibling_pos, label)
                    )
                    self.add_binary_feature(
                        'H9=%s_%s_%s' % (info.token, left_sibling_token, label)
                    )
                    self.add_binary_feature(
                        'H10=%s_%s_%s'
                        % (info.target_pos, right_sibling_pos, label)
                    )
                    self.add_binary_feature(
                        'H11=%s_%s_%s'
                        % (info.token, right_sibling_token, label)
                    )

                if use_grandparent_features:
                    self.add_binary_feature(
                        'H12=%s_%s_%d_%s'
                        % (
                            info.target_pos,
                            grandparent_pos,
                            int(grandparent_on_left),
                            label,
                        )
                    )
                    self.add_binary_feature(
                        'H13=%s_%s_%d_%s'
                        % (
                            info.token,
                            grandparent_token,
                            int(grandparent_on_left),
                            label,
                        )
                    )
                    self.add_binary_feature(
                        'H14=%s_%s_%s_%d_%s'
                        % (
                            info.target_pos,
                            head_pos,
                            grandparent_pos,
                            int(grandparent_on_left),
                            label,
                        )
                    )
                    self.add_binary_feature(
                        'H15=%s_%s_%s_%d_%s'
                        % (
                            info.token,
                            head_pos,
                            grandparent_token,
                            int(grandparent_on_left),
                            label,
                        )
                    )
                    self.add_binary_feature(
                        'H16=%s_%s_%s_%d_%s'
                        % (
                            info.token,
                            head_token,
                            grandparent_pos,
                            int(grandparent_on_left),
                            label,
                        )
                    )
                    self.add_binary_feature(
                        'H17=%s_%s_%s_%d_%s'
                        % (
                            info.target_pos,
                            head_token,
                            grandparent_token,
                            int(grandparent_on_left),
                            label,
                        )
                    )
                    self.add_binary_feature(
                        'H18=%s_%s_%s_%d_%s'
                        % (
                            info.target_pos,
                            head_pos,
                            grandparent_token,
                            int(grandparent_on_left),
                            label,
                        )
                    )
                    self.add_binary_feature(
                        'H19=%s_%s_%s_%d_%s'
                        % (
                            info.target_pos,
                            head_token,
                            grandparent_pos,
                            int(grandparent_on_left),
                            label,
                        )
                    )
                    self.add_binary_feature(
                        'H20=%s_%s_%s_%d_%s'
                        % (
                            info.token,
                            head_pos,
                            grandparent_pos,
                            int(grandparent_on_left),
                            label,
                        )
                    )

            if use_stacked_features:
                if len(info.stacked_features) > 0:
                    for i, value in enumerate(info.stacked_features):
                        self.add_numeric_feature('S%d_%s' % (i, label), value)

        if self.save_to_cache:
            self.save_cached_features()
            return

    def compute_bigram_features(self, sentence_word_features, part):
        """Compute bigram features (that depend on consecutive labels)."""
        if self.load_from_cache:
            self.load_cached_features()
            return

        index = part.index
        label = part.label
        previous_label = part.previous_label

        ignore_source = False
        only_basic_features = self.use_basic_features_only
        use_client_features = self.use_client_features
        use_parse_features = self.use_parse_features
        use_stacked_features = self.use_stacked_features  # False
        use_bias = True
        # True for using only a single bigram indicator feature.
        use_only_bias = self.use_simple_bigram_features
        use_language_model = True
        use_binary_features = False
        use_trigram_features = True

        if use_parse_features:
            use_split_morphs = False
            use_morph_features = False
            use_deprel_features = True
            use_head_features = False
            use_sibling_features = False
        else:
            use_split_morphs = False
            use_morph_features = False
            use_deprel_features = False
            use_head_features = False
            use_sibling_features = False

        if index < len(sentence_word_features):
            info = sentence_word_features[index]
        else:
            info = LinearWordQESentence.create_stop_symbol()

        if index > 0:
            info_previous = sentence_word_features[index - 1]
        else:
            info_previous = LinearWordQESentence.create_stop_symbol()

        bigram_label = str(previous_label) + '_' + str(label)
        if use_client_features:
            labels = [bigram_label, info.client_name + '_' + bigram_label]
        else:
            labels = [bigram_label]

        for label in labels:
            if use_bias:
                self.add_binary_feature('B1=%s' % label)

            if use_only_bias:
                continue

            self.add_binary_feature('B2=%s_%s' % (info.token, label))
            self.add_binary_feature('B3=%s_%s' % (info_previous.token, label))
            self.add_binary_feature('B4=%s_%s' % (info.right_context, label))
            self.add_binary_feature(
                'B5=%s_%s' % (info_previous.left_context, label)
            )

            if not ignore_source:
                self.add_binary_feature(
                    'B6=%s_%s' % (info.first_aligned_token, label)
                )
                self.add_binary_feature(
                    'B7=%s_%s' % (info.left_alignment, label)
                )
                self.add_binary_feature(
                    'B8=%s_%s' % (info.right_alignment, label)
                )
                self.add_binary_feature(
                    'B9=%s_%s' % (info_previous.first_aligned_token, label)
                )
                self.add_binary_feature(
                    'B10=%s_%s' % (info_previous.left_alignment, label)
                )
                self.add_binary_feature(
                    'B11=%s_%s' % (info_previous.right_alignment, label)
                )

            if use_binary_features and not only_basic_features:
                # Ablated for German WMT16 (the provided stoplist is wrong).
                # self.add_binary_feature(
                #     'B12=%d_%s' % (int(info.is_stopword), label))
                # self.add_binary_feature(
                #     'B13=%d_%s' % (int(info_previous.is_stopword), label))
                self.add_binary_feature(
                    'B14=%d_%s' % (int(info.is_punctuation), label)
                )
                self.add_binary_feature(
                    'B15=%d_%s' % (int(info_previous.is_punctuation), label)
                )
                # Ablated for German (capitalized words are nouns)
                # self.add_binary_feature(
                #     'B16=%d_%s' % (int(info.is_proper_noun), label))
                # self.add_binary_feature(
                #     'B17=%d_%s' % (int(info_previous.is_proper_noun), label))
                self.add_binary_feature(
                    'B18=%d_%s' % (int(info.is_digit), label)
                )
                self.add_binary_feature(
                    'B19=%d_%s' % (int(info_previous.is_digit), label)
                )

            if use_language_model and not only_basic_features:
                self.add_binary_feature(
                    'B20=%d_%s' % (info.highest_order_ngram_left, label)
                )
                self.add_binary_feature(
                    'B21=%d_%s' % (info.highest_order_ngram_right, label)
                )
                self.add_binary_feature(
                    'B22=%d_%s'
                    % (info_previous.highest_order_ngram_left, label)
                )
                self.add_binary_feature(
                    'B23=%d_%s'
                    % (info_previous.highest_order_ngram_right, label)
                )

            # if use_language_model and not only_basic_features:
            #     self.add_binary_feature(
            #         'B24=%d_%s' % (info.backoff_behavior_left, label))
            #     self.add_binary_feature(
            #         'B25=%d_%s' % (info.backoff_behavior_middle, label))
            #     self.add_binary_feature(
            #         'B26=%d_%s' % (info.backoff_behavior_right, label))
            #     self.add_binary_feature(
            #         'B27=%d_%s' % (info_previous.backoff_behavior_left,
            #                        label))
            #     self.add_binary_feature(
            #         'B28=%d_%s' % (info_previous.backoff_behavior_middle,
            #                        label))
            #     self.add_binary_feature(
            #         'B29=%d_%s' % (info_previous.backoff_behavior_right,
            #                        label))

            if use_language_model and not only_basic_features:
                self.add_binary_feature(
                    'B30=%d_%s' % (info.source_highest_order_ngram_left, label)
                )
                self.add_binary_feature(
                    'B31=%d_%s' % (info.source_highest_order_ngram_right, label)
                )
                self.add_binary_feature(
                    'B33=%d_%s'
                    % (info_previous.source_highest_order_ngram_left, label)
                )
                self.add_binary_feature(
                    'B34=%d_%s'
                    % (info_previous.source_highest_order_ngram_right, label)
                )

            if not only_basic_features:
                self.add_binary_feature('B35=%s_%s' % (info.target_pos, label))
                self.add_binary_feature(
                    'B36=%s_%s' % (info.aligned_source_pos_list, label)
                )
                self.add_binary_feature(
                    'B37=%s_%s' % (info_previous.target_pos, label)
                )
                self.add_binary_feature(
                    'B38=%s_%s' % (info_previous.aligned_source_pos_list, label)
                )

            # Conjoined features.
            self.add_binary_feature(
                'C0=%s_%s_%s' % (info.token, info.left_context, label)
            )
            self.add_binary_feature(
                'C1=%s_%s_%s' % (info.token, info.right_context, label)
            )
            self.add_binary_feature(
                'C2=%s_%s_%s'
                % (info_previous.token, info_previous.left_context, label)
            )
            self.add_binary_feature(
                'C3=%s_%s_%s'
                % (info_previous.token, info_previous.right_context, label)
            )

            if use_trigram_features:
                self.add_binary_feature(
                    'D1=%s_%s_%s_%s'
                    % (
                        info_previous.left_context,
                        info_previous.token,
                        info.token,
                        label,
                    )
                )
                self.add_binary_feature(
                    'D2=%s_%s_%s_%s'
                    % (
                        info_previous.token,
                        info.token,
                        info.right_context,
                        label,
                    )
                )

            if not ignore_source:
                self.add_binary_feature(
                    'C4=%s_%s_%s'
                    % (info.token, info.first_aligned_token, label)
                )
                self.add_binary_feature(
                    'C5=%s_%s_%s'
                    % (
                        info_previous.token,
                        info_previous.first_aligned_token,
                        label,
                    )
                )
            if not only_basic_features:
                self.add_binary_feature(
                    'C6=%s_%s_%s'
                    % (info.target_pos, info.aligned_source_pos_list, label)
                )
                self.add_binary_feature(
                    'C7=%s_%s_%s'
                    % (
                        info_previous.target_pos,
                        info_previous.aligned_source_pos_list,
                        label,
                    )
                )

            # Parse features.
            if use_parse_features:
                head_index = info.target_head - 1
                previous_head_index = info_previous.target_head - 1
                if head_index >= 0:
                    head_info = sentence_word_features[head_index]
                    head_token = head_info.token
                    head_pos = head_info.target_pos
                    head_morph = head_info.target_morph
                elif head_index == -1:
                    head_info = None
                    head_token = '__ROOT__'
                    head_pos = '__ROOT__'
                    head_morph = '__ROOT__'
                else:
                    head_info = None
                    head_token = '__START__'
                    head_pos = '__START__'
                    head_morph = '__START__'
                if previous_head_index >= 0:
                    previous_head_info = sentence_word_features[
                        previous_head_index
                    ]
                    previous_head_token = previous_head_info.token
                    previous_head_pos = previous_head_info.target_pos
                    previous_head_morph = previous_head_info.target_morph
                elif previous_head_index == -1:
                    previous_head_info = None
                    previous_head_token = '__ROOT__'
                    previous_head_pos = '__ROOT__'
                    previous_head_morph = '__ROOT__'
                else:
                    previous_head_info = None
                    previous_head_token = '__START__'
                    previous_head_pos = '__START__'
                    previous_head_morph = '__START__'

                left_sibling_token, left_sibling_pos, right_sibling_token, right_sibling_pos = self.get_siblings(  # NOQA
                    sentence_word_features, index
                )

                previous_left_sibling_token, previous_left_sibling_pos, previous_right_sibling_token, previous_right_sibling_pos = self.get_siblings(  # NOQA
                    sentence_word_features, index - 1
                )

                if use_deprel_features:
                    self.add_binary_feature(
                        'D0=%s_%s' % (info_previous.target_deprel, label)
                    )
                    self.add_binary_feature(
                        'D1=%s_%s_%s'
                        % (
                            info_previous.token,
                            info_previous.target_deprel,
                            label,
                        )
                    )

                if use_head_features:
                    self.add_binary_feature(
                        'D2=%s_%s_%s'
                        % (info_previous.target_pos, previous_head_pos, label)
                    )
                    self.add_binary_feature(
                        'D3=%s_%s_%s'
                        % (info_previous.token, previous_head_token, label)
                    )

                if use_morph_features:
                    self.add_binary_feature(
                        'D4=%s_%s' % (info_previous.target_morph, label)
                    )
                    self.add_binary_feature(
                        'D5=%s_%s_%s'
                        % (
                            info_previous.target_morph,
                            previous_head_morph,
                            label,
                        )
                    )

                if use_split_morphs:
                    all_morphs = info_previous.target_morph.split('|')
                    all_head_morphs = previous_head_morph.split('|')
                    for m in all_morphs:
                        self.add_binary_feature('D6=%s_%s' % (m, label))
                        for hm in all_head_morphs:
                            self.add_binary_feature(
                                'D7=%s_%s_%s' % (m, hm, label)
                            )

                if use_sibling_features:
                    self.add_binary_feature(
                        'D8=%s_%s_%s'
                        % (
                            info_previous.target_pos,
                            previous_left_sibling_pos,
                            label,
                        )
                    )
                    self.add_binary_feature(
                        'D9=%s_%s_%s'
                        % (
                            info_previous.token,
                            previous_left_sibling_token,
                            label,
                        )
                    )
                    self.add_binary_feature(
                        'D10=%s_%s_%s'
                        % (
                            info_previous.target_pos,
                            previous_right_sibling_pos,
                            label,
                        )
                    )
                    self.add_binary_feature(
                        'D11=%s_%s_%s'
                        % (
                            info_previous.token,
                            previous_right_sibling_token,
                            label,
                        )
                    )

                if use_deprel_features:
                    self.add_binary_feature(
                        'E0=%s_%s' % (info.target_deprel, label)
                    )
                    self.add_binary_feature(
                        'E1=%s_%s_%s' % (info.token, info.target_deprel, label)
                    )

                if use_head_features:
                    self.add_binary_feature(
                        'E2=%s_%s_%s' % (info.target_pos, head_pos, label)
                    )
                    self.add_binary_feature(
                        'E3=%s_%s_%s' % (info.token, head_token, label)
                    )

                if use_morph_features:
                    self.add_binary_feature(
                        'E4=%s_%s' % (info.target_morph, label)
                    )
                    self.add_binary_feature(
                        'E5=%s_%s_%s' % (info.target_morph, head_morph, label)
                    )

                if use_split_morphs:
                    all_morphs = info.target_morph.split('|')
                    all_head_morphs = head_morph.split('|')
                    for m in all_morphs:
                        self.add_binary_feature('E6=%s_%s' % (m, label))
                        for hm in all_head_morphs:
                            self.add_binary_feature(
                                'E7=%s_%s_%s' % (m, hm, label)
                            )

                if use_sibling_features:
                    self.add_binary_feature(
                        'E8=%s_%s_%s'
                        % (info.target_pos, left_sibling_pos, label)
                    )
                    self.add_binary_feature(
                        'E9=%s_%s_%s' % (info.token, left_sibling_token, label)
                    )
                    self.add_binary_feature(
                        'E10=%s_%s_%s'
                        % (info.target_pos, right_sibling_pos, label)
                    )
                    self.add_binary_feature(
                        'E11=%s_%s_%s'
                        % (info.token, right_sibling_token, label)
                    )
            if use_stacked_features:
                if len(info.stacked_features) > 0:
                    for i, value in enumerate(info.stacked_features):
                        self.add_numeric_feature('Z%d_%s' % (i, label), value)
                if len(info_previous.stacked_features) > 0:
                    for i, value in enumerate(info_previous.stacked_features):
                        self.add_numeric_feature('ZZ%d_%s' % (i, label), value)

        if self.save_to_cache:
            self.save_cached_features()
            return



#  OpenKiwi: Open-Source Machine Translation Quality Estimation
#  Copyright (C) 2019 Unbabel <openkiwi@unbabel.com>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import logging

from kiwi.cli.better_argparse import ModelParser
from kiwi.cli.opts import PathType
from kiwi.models.linear_word_qe_classifier import LinearWordQEClassifier

logger = logging.getLogger(__name__)

title = 'linear'


def _add_vocabulary_opts(parser):
    group = parser.add_argument_group('vocabulary options')
    group.add_argument(
        '--source-vocab-size',
        type=int,
        default=None,
        help='Size of the source vocabulary.',
    )
    group.add_argument(
        '--target-vocab-size',
        type=int,
        default=None,
        help='Size of the target vocabulary.',
    )
    group.add_argument(
        '--source-vocab-min-frequency',
        type=int,
        default=1,
        help='Min word frequency for source vocabulary.',
    )
    group.add_argument(
        '--target-vocab-min-frequency',
        type=int,
        default=1,
        help='Min word frequency for target vocabulary.',
    )


def add_training_data_file_opts(parser):
    # Data options
    group = parser.add_argument_group('data')

    group.add_argument(
        '--train-source',
        type=PathType(exists=True),
        help='Path to training source file',
    )
    group.add_argument(
        '--train-target',
        type=PathType(exists=True),
        help='Path to training target file',
    )
    group.add_argument(
        '--train-alignments',
        type=str,
        help='Path to train alignments between source and target.',
    )
    group.add_argument(
        '--train-source-tags',
        type=PathType(exists=True),
        help='Path to validation label file for source (WMT18 format)',
    )
    group.add_argument(
        '--train-target-tags',
        type=PathType(exists=True),
        help='Path to validation label file for target',
    )

    group.add_argument(
        '--train-source-pos',
        type=PathType(exists=True),
        help='Path to training PoS tags file for source',
    )
    group.add_argument(
        '--train-target-pos',
        type=PathType(exists=True),
        help='Path to training PoS tags file for target',
    )
    group.add_argument(
        '--train-target-parse',
        type=PathType(exists=True),
        help='Path to training dependency parsing file for target (tabular '
        'format)',
    )
    group.add_argument(
        '--train-target-ngram',
        type=PathType(exists=True),
        help='Path to training highest order ngram file for target (tabular '
        'format)',
    )
    group.add_argument(
        '--train-target-stacked',
        type=PathType(exists=True),
        help='Path to training stacked predictions file for target (tabular '
        'format)',
    )

    group = parser.add_argument_group('validation data')

    group.add_argument(
        '--valid-source',
        type=PathType(exists=True),
        # required=True,
        help='Path to validation source file',
    )
    group.add_argument(
        '--valid-target',
        type=PathType(exists=True),
        # required=True,
        help='Path to validation target file',
    )
    group.add_argument(
        '--valid-alignments',
        type=str,
        # required=True,
        help='Path to valid alignments between source and target.',
    )
    group.add_argument(
        '--valid-source-tags',
        type=PathType(exists=True),
        help='Path to validation label file for source (WMT18 format)',
    )
    group.add_argument(
        '--valid-target-tags',
        type=PathType(exists=True),
        help='Path to validation label file for target',
    )

    group.add_argument(
        '--valid-source-pos',
        type=PathType(exists=True),
        help='Path to training PoS tags file for source',
    )
    group.add_argument(
        '--valid-target-pos',
        type=PathType(exists=True),
        help='Path to training PoS tags file for target',
    )
    group.add_argument(
        '--valid-target-parse',
        type=PathType(exists=True),
        help='Path to validation dependency parsing file for target (tabular '
        'format)',
    )
    group.add_argument(
        '--valid-target-ngram',
        type=PathType(exists=True),
        help='Path to validation highest order ngram file for target (tabular '
        'format)',
    )
    group.add_argument(
        '--valid-target-stacked',
        type=PathType(exists=True),
        help='Path to validation stacked predictions file for target (tabular '
        'format)',
    )


def add_predicting_data_file_opts(parser):
    # Data options
    group = parser.add_argument_group('data')

    group.add_argument(
        '--test-source',
        type=PathType(exists=True),
        required=True,
        help='Path to validation source file',
    )
    group.add_argument(
        '--test-target',
        type=PathType(exists=True),
        required=True,
        help='Path to validation target file',
    )
    group.add_argument(
        '--test-alignments',
        type=PathType(exists=True),
        help='Path to test alignments between source and target.',
    )
    group.add_argument(
        '--test-source-pos',
        type=PathType(exists=True),
        help='Path to training PoS tags file for source',
    )
    group.add_argument(
        '--test-target-pos',
        type=PathType(exists=True),
        help='Path to training PoS tags file for target',
    )
    group.add_argument(
        '--test-target-parse',
        type=PathType(exists=True),
        help='Path to test dependency parsing file for target (tabular format)',
    )
    group.add_argument(
        '--test-target-ngram',
        type=PathType(exists=True),
        help='Path to test highest order ngram file for target (tabular '
        'format)',
    )  # noqa
    group.add_argument(
        '--test-target-stacked',
        type=PathType(exists=True),
        help='Path to test stacked predictions file for target (tabular '
        'format)',
    )  # noqa

    return group


def _add_output_options(group):
    # Other options (used both at training and test time).
    group.add_argument(
        '--evaluation-metric',
        type=str,
        default='f1_mult',
        help='Evaluation metric (f1_mult or f1_bad).',
    )


def add_training_options(training_parser):
    add_training_data_file_opts(training_parser)
    _add_vocabulary_opts(training_parser)

    group = training_parser.add_argument_group(
        'linear', description='Linear Quality Estimation'
    )

    # Model options (training time).
    group.add_argument(
        '--use-basic-features-only',
        type=int,
        default=0,
        help='1 for using only basic features (words).',
    )
    group.add_argument(
        '--use-bigrams',
        type=int,
        default=1,
        help='1 for using bigram features (i.e. a CRF-like model).',
    )
    group.add_argument(
        '--use-simple-bigram-features',
        type=int,
        default=0,
        help='1 for using only label indicators as bigram features.',
    )

    # Training options.
    group.add_argument(
        '--training-algorithm',
        type=str,
        default='svm_mira',
        help='Algorithm for training the model (svm_mira, svm_sgd, '
        'perceptron).',
    )
    group.add_argument(
        '--regularization-constant',
        type=float,
        default=0.001,
        help='L2 regularization constant.',
    )
    group.add_argument(
        '--cost-false-positives',
        type=float,
        default=0.2,
        help='Cost for false positives (svm_mira and svm_sgd only).',
    )
    group.add_argument(
        '--cost-false-negatives',
        type=float,
        default=0.8,
        help='Cost for false negatives (svm_mira and svm_sgd only).',
    )

    _add_output_options(group)


def add_predicting_options(predicting_parser):
    add_predicting_data_file_opts(predicting_parser)
    _add_output_options(predicting_parser)


def parser_for_pipeline(pipeline):
    if pipeline == 'train':
        return ModelParser(
            'linear',
            'train',
            title=LinearWordQEClassifier.title,
            options_fn=add_training_options,
            api_module=LinearWordQEClassifier,
        )
    if pipeline == 'predict':
        return ModelParser(
            'linear',
            'predict',
            title=LinearWordQEClassifier.title,
            options_fn=add_predicting_options,
            api_module=LinearWordQEClassifier,
        )

    return None
